unit module_defs;

{

   COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Math, System.Classes,

  KnobsUtils, KnobsConversions;

const

  NO_INPUT          = 'noinput';
  NO_OUTPUT         = 'nooutput';
  NO_INTERNAL       = 'nointernal';
  NO_STRING         = 'nostring';
  NO_MODULE         = 'nomodule';
  UNDEF_PMC         = -1;                            // Pin/module/connection index for non existent pin/module/connection


  {$IFDEF DEBUG}
    UNDEFINED = NAN;                               // NAN for stability checks
    {$INLINE OFF}
  {$ELSE}
    UNDEFINED = 0.0;                               // 0.0 for production code
    {$INLINE ON}
  {$ENDIF}

  WhiteSpace         = [ #$00 .. #$20, #$7f];
  StrictDigits       = [ '0' .. '9'];
  Digits             = StrictDigits + [ '+', '-'];
  IdentifierStarters = [ 'A' .. 'Z', 'a' .. 'z', '%', '#', '@', '_'];
  IdentifierChars    = IdentifierStarters + Digits + [ '&'];
  FloatChars         = Digits + [ '.', 'e', 'E'];


  // Midi command definitions

  CNoteOff         = $80; // Length = 3;
  CNoteOn          = $90; // Length = 3;
  CKeyPressure     = $a0; // Length = 3;
  CControlChange   = $b0; // Length = 3;
  CProgramChange   = $c0; // Length = 2;
  CChannelPressure = $d0; // Length = 2;
  CPitchBend       = $e0; // Length = 2;
  CSysEx           = $f0; // Length = n;

  // When aCmd And $f0 = CSysEx low nibble of 1st byte is :

  SSysEx           = $00; // Length = 1, followed by arbitrary amt of data, finalized by SEox;
  SSongPosition    = $02; // Length = 3;
  SSongSelect      = $03; // Length = 2;
  STuneRequest     = $06; // Length = 1;
  SEox             = $07; // Length = 1;
  STimingClock     = $08; // Length = 1;
  SStart           = $0a; // Length = 1;
  SContinue        = $0b; // Length = 1;
  SStop            = $0c; // Length = 1;
  SActiveSense     = $0e; // Length = 1;
  SReset           = $0f; // Length = 1;

  CCAllNotesOff    = 123; // Pseudo CC - all notes off

type

  EModule            = class( Exception);
  EModuleNotFound    = class( EModule  );
  EInputNotFound     = class( EModule  );
  EOutputNotFound    = class( EModule  );
  EInternalNotFound  = class( EModule  );
  EStringNotFound    = class( EModule  );
  EUnitsUndefined    = class( EModule  );

  TStringArray       = array of string;
  TConnection        = class;
  TConnectionArray   = array of TConnection;
  TMod               = class;
  TSynthPatch        = class;
  TModuleArray       = array of TMod;
  TModuleClass       = class of TMod;

  TInts              = array of Integer;
  TKnobMap           = set   of Byte;

  TMidiMessage = record
    Command : Byte;
    Channel : Byte;
    Data    : TInts;
  end;

  TLightsHandler     = procedure( aSender: TMod; const aMsg: string)                             of object;
  TDataHandler       = procedure( aSender: TMod; const aPath: string; const aData: TSignalArray) of object;
  TInfoHandler       = procedure( aSender: TMod; const aMsg: string)                             of object;
  TOnSendMidi        = procedure( aSender: TObject; const aMsg: TMidiMessage)                    of object;

  TConnection = class
  public
    Peer      : TSynthPatch;
    Index     : Integer;
    SrcModule : Integer;
    SrcPin    : Integer;
    DstModule : Integer;
    DstPin    : Integer;
  private
    procedure   Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
  public
    procedure   Save( anIndent: Integer; var S: string);
    function    IsSameConnection( aConnection: TConnection): Boolean;
  end;

  TConnections = class
  private
    FPeer       : TSynthPatch;
    FConnections: TConnectionArray;
  private
    function    GetCount: Integer;
    function    GetConnection( anIndex: Integer): TConnection;
    procedure   SetConnection( anIndex: Integer; const aValue: TConnection);
  public
    constructor Create( aPeer: TSynthPatch);
    destructor  Destroy;                                                                     override;
    procedure   Save( anIndent: Integer; const aName: string; var S: string);
    procedure   Clear;
    procedure   AddConnection( const aValue: TConnection);                                   overload;
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);           overload;
    procedure   RemoveConnection( anIndex: Integer);
    procedure   RemoveConnectionsFor( aModuleIndex: Integer);
    procedure   PropagateSignals( const aModules: TModuleArray);
    function    IsSameConnections( const aConnections: TConnections): Boolean;
  public
    property    Count: Integer read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  end;

  TMod = class
  private
    FParent      : TSynthPatch;
    FName        : string;
    FInputs      : TSignals;
    FShadows     : TSignals;
    FOutputs     : TSignals;
    FResetFlag   : Boolean;
    FSamples     : Int64;
    FClocks      : Int64;
    FTotalClocks : Int64;
    FZips        : Int64;
    FTotalZips   : Int64;
  protected
    FIsFast      : Boolean;           // True when fast mode processing must be performed
    FIsMidi      : Boolean;           // True when midi reception must be enabled
    FInCache     : TSignalArray;
    FShadowCache : TSignalArray;
    FOutCache    : TSignalArray;
    FKnobMap     : TKnobMap;
  private
    function    AssertUniqueName( const aName: string): string;                                   virtual;
  private
    procedure   SetName( const aValue: string);
    function    GetInputCount  : Integer;
    function    GetOutputCount : Integer;
    function    GetOutput( anIndex: Integer): TSignal;
    procedure   SetCommonValue        ( anIndex : Integer; aValue: TSignal);                      virtual;
    procedure   SetInternalValue      ( const aName: string; aValue: TSignal);                    virtual;
    procedure   SetInternalStringValue( const aName, aValue: string);
  protected
    procedure   SetStringValue        ( const aName, aValue: string);                             virtual;
    function    GetStringValue        ( const aName: string): string;                             virtual;
    procedure   FixInputs;
    procedure   FixOutputs;
    procedure   SetInputs( aStartIndex: Integer; const aValue: TSignals);
    procedure   HandleParam( const aControl: string; aValue: TSignal);                            virtual;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                               virtual;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                      virtual;
    procedure   SaveInputValues ( anIndent: Integer; var S: string);
    procedure   SaveKeyValue    ( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    procedure   SaveDetails     ( anIndent: Integer; var S: string);                              virtual;
    procedure   SaveInternals   ( anIndent: Integer; var S: string);                              virtual;
    procedure   SaveStrings     ( anIndent: Integer; var S: string);                              virtual;
    procedure   SetInternal     ( const aName: string; aValue: TSignal);                          virtual;
    function    GetInternal     ( const aName: string): TSignal;                                  virtual;
    procedure   Save            ( anIndent: Integer; var S: string);
    procedure   Reset;                                                                            virtual;
    procedure   Clear;                                                                            virtual;
    function    FindInput   ( const aPinName: string): Integer;
    function    FindOutput  ( const aPinName: string): Integer;
    function    InputName   ( anIndex: Integer): string;                                          virtual;
    function    OutputName  ( anIndex: Integer): string;                                          virtual;
    function    InternalName( anIndex: Integer): string;                                          virtual;
    function    StringName  ( anIndex: Integer): string;                                          virtual;
    procedure   PreTick;                                                                           inline;
    procedure   PostTick;                                                                          inline;
    procedure   PreZipper;                                                                         inline;
    procedure   PostZipper;                                                                        inline;
    procedure   Tick;
    procedure   SlowTick;
    procedure   DeZipper;
    procedure   DoTick;                                                                           virtual;
    procedure   DoSlowTick;                                                                       virtual;
    procedure   DoDeZipper;
    procedure   FixZippers;                                                                       virtual;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);            virtual;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);              virtual;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);              virtual;
    procedure   AcceptParam( const aPath: string; aValue: TSignal);
    function    GetProfileInfoHtml: string;
    function    IsSameModule( const aModule: TMod): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                            virtual;
    procedure   SendMidi( const aMsg: TMidiMessage);                                              virtual;
  public
    property    Name                             : string   read FName                  write SetName;
    property    ResetFlag                        : Boolean  read FResetFlag             write FResetFlag;
    property    TotalClocks                      : Int64    read FTotalClocks;
    property    TotalZips                        : Int64    read FTotalZips;
    property    Samples                          : Int64    read FSamples;
    property    InputCount                       : Integer  read GetInputCount;
    property    OutputCount                      : Integer  read GetOutputCount;
    property    Output[ anIndex: Integer]        : TSignal  read GetOutput;
  public
    property    InternalValue      [ const aName: string] : TSignal                     write SetInternalValue;
    property    InternalStringValue[ const aName: string] : string                      write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string  read GetStringValue write SetStringValue;
  end;

  TModInput = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModOutput = class( TMod)
  public
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TSynthPatch = class( TMod)
  private
    FVersion          : Integer;
    FModules          : TModuleArray;
    FConnections      : TConnections;
    FDecimatorCount   : Integer;
    FInputQueue       : TSignalQueue;
    FWireClocks       : Int64;
    FTotalWireClocks  : Int64;
    FOnSendMidi       : TOnSendMidi;
  private
    function    AssertUniqueName( const aName: string): string;                                  override;
  private
    function    CheckModuleIndex( anIndex: Integer): Boolean;
    function    GetModuleCount: Integer;
    function    GetModule( anIndex: Integer): TMod;
    function    GetConnectionCount: Integer;
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
  protected
    procedure   HandleParam( const aPath: string; aValue: TSignal);                              override;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SaveConnections   ( anIndent: Integer; var S: string);
    procedure   SaveModules       ( anIndent: Integer; var S: string);
    procedure   SaveDetails       ( anIndent: Integer; var S: string);                           override;
    procedure   Clear;                                                                           override;
    procedure   Reset;                                                                           override;
    procedure   RemoveConnectionsFor( aModuleIndex: Integer);
    procedure   ClearModules;
    function    FindModule( const aModuleName: string): Integer;
    function    ModuleName( anIndex: Integer): string;
    function    FindInputPin ( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    InputPinName ( aModuleIndex, aPinIndex: Integer): string;
    function    OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    procedure   AddModule( const aModule: TMod);
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);               overload;
    procedure   AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string);          overload;
    procedure   PreWires;
    procedure   PostWires;
    procedure   PropagateSignals;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    procedure   DoTick;                                                                          override;
    procedure   FixZippers;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    procedure   AcceptSamplePair( aLeft, aRight: TSignal);
    function    GetProfileHtml: TStringList;
    function    IsSamePatch       ( const aPatch: TSynthPatch): Boolean;
    function    SameModules       ( const aPatch: TSynthPatch): Boolean;
    function    SameConnections   ( const aPatch: TSynthPatch): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   SendMidi( const aMsg: TMidiMessage);                                             override;
  public
    property    Version                      : Integer     read FVersion       write FVersion;
    property    ModuleCount                  : Integer     read GetModuleCount;
    property    Module    [ anIndex: Integer]: TMod        read GetModule;
    property    ConnectionCount              : Integer     read GetConnectionCount;
    property    Overflows                    : Integer     read GetOverflows;
    property    Underflows                   : Integer     read GetUnderflows;
    property    OnSendMidi                   : TOnSendMidi read FOnSendMidi    write FOnSendMidi;
  end;


  procedure ParsePrefix( const aPath: string; var aPrefix, aRest: string);
  function  ParsePostfix( const aPath: string): string;
  function  ApplyUnitsToSignal( var aValue: TSignal; const aUnits: string): Boolean;
  function  ParseTextValue( const aString: string; var aSignal: TSignal): Boolean;

  function  ParseTextSequencerValue( const aValue: string): TSignals;

  function  MakeMidiNoteOff        ( aVelocity, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiNoteOn         ( aVelocity, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiKeyPressure    ( aPressure, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiControlChange  ( aValue, aController, aChannel: Byte) : TMidiMessage;
  function  MakeMidiProgramChange  ( aProgram, aChannel: Byte)            : TMidiMessage;
  function  MakeMidiChannelPressure( aValue, aChannel: Byte)              : TMidiMessage;
  function  MakeMidiPitchBend      ( aData: Word; aChannel: Byte)         : TMidiMessage;
  function  MakeMidiSongPosition   ( aData: Word)                         : TMidiMessage;
  function  MakeMidiSongSelect     ( aSong: Byte)                         : TMidiMessage;
  function  MakeMidiTuneRequest                                           : TMidiMessage;
  function  MakeMidiTimingClock                                           : TMidiMessage;
  function  MakeMidiStart                                                 : TMidiMessage;
  function  MakeMidiContinue                                              : TMidiMessage;
  function  MakeMidiStop                                                  : TMidiMessage;
  function  MakeMidiActiveSense                                           : TMidiMessage;
  function  MakeMidiReset                                                 : TMidiMessage;
  function  MakeMidiSysExData      ( const aData: TBytes)                 : TMidiMessage;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  function  MidiMessageToLog( const aMsg: TMidiMessage): string;



implementation



uses

  Globals;


  procedure ParsePrefix( const aPath: string; var aPrefix, aRest: string);
  // NOTE: this code needs be compiled with "String Format Checking" on,
  //       otherwise it will generate a run time error.
  //       ok .. this will only occur when aPath and aRest refer to the same
  //       underlying string .. as in ParsePrefix( aRest, aPrefix, aRest);
  var
    i  : Integer;
  begin
    i       := 1;
    aPrefix := '';
    aRest   := '';
    while i <= Length( aPath)
    do begin
      if aPath[ i] = '/'
      then begin
        aPrefix := Copy( aPath, 1, i - 1 );                    // Either on this line
        aRest := Copy( aPath, i + 1, Length( aPath));          // Or this one .. not quite clear.
        Break;
      end
      else Inc( i);
    end;
  end;

  function  ParsePostfix( const aPath: string): string;
  var
    i : Integer;
  begin
    Result := '';
    for i := Length( aPath) downto 1
    do begin
      if aPath[ i] = '_'
      then begin
        Result := Copy( aPath, i + 1, Length( aPath));
        Break;
      end
    end;
  end;

  function  ApplyUnitsToSignal( var aValue: TSignal; const aUnits: string): Boolean;
  begin
    Result := True;
    if      SameText( aUnits, '')
    then // nothing
    else if SameText( aUnits, 'Hz')
    then aValue := FrequencyToUnits( aValue)
    else if SameText( aUnits, 'kHz')
    then aValue := FrequencyToUnits( 1000 * aValue)
    else if SameText( aUnits, 'mHz')
    then aValue := FrequencyToUnits( aValue / 1000)
    else if SameText( aUnits, 'bpm')
    then aValue := FrequencyToUnits( aValue / 60.0)
    else if SameText( aUnits, 'degrees') or SameText( aUnits, 'deg')
    then aValue := aValue / 360
    else if SameText( aUnits, 'radians') or SameText( aUnits, 'rad')
    then aValue := aValue / ( PI_2)
    else if SameText( aUnits, 'percent') or SameText( aUnits, '%')
    then aValue := aValue / 100
    else if SameText( aUnits, 'ms')
    then aValue := aValue / 1000
    else if SameText( aUnits, 's')
    then // nothing
    else if SameText( aUnits, 'nn')
    then aValue := NoteNumberToUnits( aValue)
    else if SameText( aUnits, 'qsteps') or SameText( aUnits, 'q')
    then aValue := aValue / 1000
    else Result := False;
  end;

  function  StripWhiteSpace( const S: string): string;
  var
    C: Char;
  begin
    Result := '';
    for C in S
    do begin
      if not CharInSet( C, WhiteSpace)
      then Result := Result + C;
    end;
  end;

  function  ParseTextValue( const aString: string; var aSignal: TSignal): Boolean;
  var
    aParts  : TStringList;
    S       : string;
    aUnits  : string;
  begin
    Result := True;
    aParts := Explode( aString, ' ');
    try
      aUnits := '';
      if aParts.Count > 0
      then begin
        S := Trim( aParts[ 0]);
        if IsNoteName( S)
        then aSignal := NoteNameToUnits( S, UNDEFINED)
        else aSignal := StrToFloatDef  ( S, UNDEFINED);
      end
      else Result := False;
      if Result and ( aParts.Count > 1)
      then begin
        aUnits := Trim( aParts[ 1]);
        Result := ApplyUnitsToSignal( aSignal, aUnits);
      end;
    finally
      aParts.Free;
    end;
  end;

  function  ParseTextSequencerValue( const aValue: string): TSignals;
  var
    aStrings : TStringList;
    i        : Integer;
    aSignal  : TSignal;
  begin
    Result   := TSignals.Create( 0);
    aStrings := Explode( StripWhiteSpace( aValue), ',');
    try
      for i := 0 to aStrings.Count - 1
      do begin
        if ParseTextValue( aStrings[ i], aSignal)
        then Result.AddSignal( aSignal);
      end;
    finally
      aStrings.Free;
    end;
  end;

  function  MakeMidiNoteOff( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOff;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiNoteOn( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOn;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiKeyPressure( aPressure, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CKeyPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aPressure;
  end;

  function  MakeMidiControlChange( aValue, aController, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CControlChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aController;
    Result.Data[ 1] := aValue;
  end;

  function  MakeMidiProgramChange( aProgram, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CProgramChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aProgram;
  end;

  function  MakeMidiChannelPressure( aValue, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CChannelPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aValue;
  end;

  function  MakeMidiPitchBend( aData: Word; aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CPitchBend;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongPosition( aData: Word): TMidiMessage;
  begin
    Result.Command := SSongPosition;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongSelect( aSong: Byte): TMidiMessage;
  begin
    Result.Command := SSongSelect;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aSong;
  end;

  function  MakeMidiTuneRequest: TMidiMessage;
  begin
    Result.Command := STuneRequest;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiTimingClock: TMidiMessage;
  begin
    Result.Command := STimingClock;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStart: TMidiMessage;
  begin
    Result.Command := SStart;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiContinue: TMidiMessage;
  begin
    Result.Command := SContinue;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStop: TMidiMessage;
  begin
    Result.Command := SStop;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiActiveSense: TMidiMessage;
  begin
    Result.Command := SActiveSense;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiReset: TMidiMessage;
  begin
    Result.Command := SReset;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiSysExData( const aData: TBytes): TMidiMessage;
  var
    i : Integer;
  begin
    Result.Command := SSysEx;
    Result.Channel := 0;
    SetLength( Result.Data, Length( aData));
    for i := 0 to Length( Result.Data) - 1
    do Result.Data[ i] := aData[ i];
  end;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  begin
    SetLength( aData, 0);

    case aMsg.Command of

      CNoteOff,          // = $80; // Length = 3;
      CNoteOn,           // = $90; // Length = 3;
      CKeyPressure,      // = $a0; // Length = 3;
      CControlChange :   // = $b0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
            aData[ 2] := aMsg.Data[ 1];
          end;
        end;

      CProgramChange,    // = $c0; // Length = 2;
      CChannelPressure : // = $d0; // Length = 2;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 2);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
          end;
        end;

      CPitchBend :       // = $e0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command  and $00f0) or ( aMsg.Channel and $000f);
            aData[ 1] :=   aMsg.Data[ 0] and $007f;
            aData[ 2] := ( aMsg.Data[ 0] and $3f80) shr 7;
          end;
        end;

//      CSysEx :          // = $f0; // Length = n;
//        begin
//        end;

    end;
  end;

  function  MidiMessageToLog( const aMsg: TMidiMessage): string;

    function CmdToStr( aCmd: Byte): string;
    begin
      case aCmd of
        CNoteOff         : Result := 'note off';         // = $80; // Length = 3;
        CNoteOn          : Result := 'note on';          // = $90; // Length = 3;
        CKeyPressure     : Result := 'key pressure';     // = $a0; // Length = 3;
        CControlChange   : Result := 'control change';   // = $b0; // Length = 3;
        CProgramChange   : Result := 'program change';   // = $c0; // Length = 2;
        CChannelPressure : Result := 'channel pressure'; // = $d0; // Length = 2;
        CPitchBend       : Result := 'pitch bend';       // = $e0; // Length = 2;
        CSysEx           : Result := 'sysex';            // = $f0; // Length = n;
      end;
    end;

  var
    i : Integer;
  begin
    Result := Format( 'ch %d cmd %.2x [%s] ', [ aMsg.Channel + 1, aMsg.Command, CmdToStr( aMsg.Command)]);
    for i := 0 to Length( aMsg.Data) - 1
    do Result := Result + Format( '%.2x ', [ aMSg.Data[ i]]);
    SetLength( Result, Length( Result) - 1);
  end;


{ ========
  TConnection = class
  private
    FPeer      : TPatch;
    FIndex     : Integer;
    FSrcModule : Integer;
    FSrcPin    : Integer;
    FDstModule : Integer;
    FDstPin    : Integer;
  public
    property    SrcModule: Integer read FSrcModule;
    property    SrcPin   : Integer read FSrcPin;
    property    DstModule: Integer read FDstModule;
    property    DstPin   : Integer read FDstPin;
  private
}

    procedure   TConnection.Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
    begin
      Peer      := aPeer;
      Index     := anIndex;
      SrcModule := aSrcModule;
      SrcPin    := aSrcPin;
      DstModule := aDstModule;
      DstPin    := aDstPin;
    end;

//  public

    procedure   TConnection.Save( anIndent: Integer; var S: string);
    var
      aSrcModName : string;
      aSrcPinName : string;
      aDstModName : string;
      aDstPinName : string;
      T           : string;
    begin
      if Assigned( Peer)
      then begin
        aSrcModName := Peer.ModuleName   ( SrcModule);
        aSrcPinName := Peer.OutputPinName( SrcModule, SrcPin);
        aDstModName := Peer.ModuleName   ( DstModule);
        aDstPinName := Peer.InputPinName ( DstModule, DstPin);

        T := Format( '%sconnect( %s %s %s %s)' + sLineBreak, [ Indent( anIndent), aSrcModName, aSrcPinName, aDstModName, aDstPinName], AppLocale);
        S := S + T
      end;
    end;

    function    TConnection.IsSameConnection( aConnection: TConnection): Boolean;
    begin
      Result := False;
      if Assigned( aConnection)
      then
        Result :=
          ( SrcModule = aConnection.SrcModule) and
          ( SrcPin    = aConnection.SrcPin   ) and
          ( DstModule = aConnection.DstModule) and
          ( DstPin    = aConnection.DstPin   );
    end;


{ ========
  TConnections = class
  private
    FConnections: TConnectionArray;
  public
    property    Count: Integer read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  private
}

    function    TConnections.GetCount: Integer;
    begin
      Result := Length( FConnections);
    end;

    function    TConnections.GetConnection( anIndex: Integer): TConnection;
    begin
      Result := FConnections[ anIndex]
    end;

    procedure   TConnections.SetConnection( anIndex: Integer; const aValue: TConnection);
    begin
      FConnections[ anIndex] := aValue;
    end;

//  public

    constructor TConnections.Create( aPeer: TSynthPatch);
    begin
      inherited Create;
      FPeer := aPeer;
    end;

    destructor  TConnections.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TConnections.Save( anIndent: Integer; const aName: string; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + aName + '(' + sLineBreak;
      for i := 0 to Count - 1
      do Connection[ i].Save( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TConnections.Clear;
    begin
      while Count > 0
      do RemoveConnection( Count - 1);
    end;

    procedure   TConnections.AddConnection( const aValue: TConnection); // overload;
    begin
      SetLength( FConnections, Count + 1);
      FConnections[ Count - 1] := aValue;
    end;

    procedure   TConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);  // overload;
    var
      aConnection: TConnection;
    begin
      aConnection := TConnection.Create;
      aConnection.Connect( FPeer, Count, aSrcModule, aSrcPin, aDstModule, aDstPin);
      AddConnection( aConnection);
    end;

    procedure   TConnections.RemoveConnection( anIndex: Integer); // overload;
    var
      aConnection: TConnection;
      i          : Integer;
    begin
      aConnection := Connection[ anIndex];
      if Assigned( aConnection)
      then begin
        for i := anIndex to Count - 2
        do Connection[ i] := Connection[ i + 1];
        SetLength( FConnections, Count - 1);
        aConnection.Free;
      end;
    end;

    procedure   TConnections.RemoveConnectionsFor( aModuleIndex: Integer);
    var
      i : Integer;
      C: TConnection;
    begin
      for i := Count - 1 downto 0
      do begin
        C := Connection[ i];
        if ( C.SrcModule = aModuleIndex) or ( C.DstModule = aModuleIndex)
        then RemoveConnection( i);
      end;
    end;

    procedure   TConnections.PropagateSignals( const aModules: TModuleArray);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do begin
        with Connection[ i]
        do aModules[ DstModule].FInCache[ DstPin] := aModules[ SrcModule].FOutCache[ SrcPin];
      end;
    end;

    function    TConnections.IsSameConnections( const aConnections: TConnections): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      if Assigned( aConnections) and ( Count = aConnections.Count)
      then begin
        Result := True;
        for i := 0 to Count - 1
        do begin
          if not Connection[ i].IsSameConnection( aConnections.Connection[ i])
          then begin
            Result := False;
            Break;
          end;
        end;
      end;
    end;


{ ========
  TMod = class
  private
    FParent      : TSynthPatch;
    FName        : string;
    FInputs      : TSignals;
    FShadows     : TSignals;
    FOutputs     : TSignals;
    FResetFlag   : Boolean;
    FSamples     : Int64;
    FClocks      : Int64;
    FTotalClocks : Int64;
    FZips        : Int64;
    FTotalZips   : Int64;
  protected
    FIsFast      : Boolean;           // True when fast mode processing must be performed
    FIsMidi      : Boolean;           // True when midi reception must be enabled
    FInCache     : TSignalArray;
    FShadowCache : TSignalArray;
    FOutCache    : TSignalArray;
    FKnobMap     : TKnobMap;
  public
    property    Name                             : string   read FName                  write SetName;
    property    ResetFlag                        : Boolean  read FResetFlag             write FResetFlag;
    property    TotalClocks                      : Int64    read FTotalClocks;
    property    TotalZips                        : Int64    read FTotalZips;
    property    Samples                          : Int64    read FSamples;
    property    InputCount                       : Integer  read GetInputCount;
    property    OutputCount                      : Integer  read GetOutputCount;
    property    Output[ anIndex: Integer]        : TSignal  read GetOutput;
  public
    property    InternalValue      [ const aName: string] : TSignal                     write SetInternalValue;
    property    InternalStringValue[ const aName: string] : string                      write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string  read GetStringValue write SetStringValue;
  private
}

    function    TMod.AssertUniqueName( const aName: string): string; // virtual;
    begin
      if Assigned( FParent)
      then Result := FParent.AssertUniqueName( aName)
      else Result := aName;
    end;

//  private

    procedure   TMod.SetName( const aValue: string);
    begin
      if aValue <> FName
      then FName := AssertUniqueName( aValue);
    end;

    function    TMod.GetInputCount: Integer;
    begin
      Result := Length( FInCache);
    end;

    function    TMod.GetOutputCount: Integer;
    begin
      Result := Length( FOutCache);
    end;

    function    TMod.GetOutput( anIndex: Integer): TSignal;
    begin
      if ( anIndex >= 0) and ( anIndex < OutputCount)
      then Result := FOutCache[ anIndex]
      else Result := 0.0;
    end;

    procedure   TMod.SetCommonValue( anIndex: Integer; aValue: TSignal); // virtual;
    begin
      if anIndex >= 0
      then begin
        if anIndex in FKnobMap
        then FShadowCache[ anIndex] := aValue
        else FInCache    [ anIndex] := aValue;
      end;
    end;

    procedure   TMod.SetInternalValue( const aName: string; aValue: TSignal); // virtual;
    var
      anIndex: Integer;
    begin
      anIndex := FindInput( aName);
      if anIndex >= 0
      then SetCommonValue( anIndex, aValue)
      else SetInternal( aName, aValue);
    end;

    procedure   TMod.SetInternalStringValue( const aName, aValue: string);
    var
      aSignal : TSignal;
    begin
      if ParseTextValue( aName, aSignal)
      then InternalValue[ aName] := aSignal
      else raise EUnitsUndefined.CreateFmt( 'Unparsable value ''%s''', [ aName])
    end;

//  protected

    procedure   TMod.SetStringValue( const aName, aValue: string); // virtual;
    begin
      // Nothing here.
    end;

    function    TMod.GetStringValue( const aName: string): string; // virtual;
    begin
      Result := '';
    end;

    procedure   TMod.FixInputs;
    var
      aCount : Integer;
    begin
      aCount := 0;
      while not SameText( InputName( aCount), NO_INPUT)
      do begin
        Inc( aCount);
        if aCount > 256
        then raise EInputNotFound.Create( 'are there really over 256 inputs? (probably a GetInputName() is broken and never returns NO_INPUT)');
      end;
      FInputs .Clear;
      FShadows.Clear;
      FInCache     := FInputs .FSignals;
      FShadowCache := FShadows.FSignals;
      while InputCount < aCount
      do begin
        FInputs .AddSignal( 0.0);
        FShadows.AddSignal( 0.0);
        FInCache     := FInputs .FSignals;
        FShadowCache := FShadows.FSignals;
      end;
    end;

    procedure   TMod.FixOutputs;
    var
      aCount : Integer;
    begin
      aCount := 0;
      while not SameText( OutputName( aCount), NO_OUTPUT)
      do begin
        Inc( aCount);
        if aCount > 256
        then raise EOutputNotFound.Create( 'are there really over 256 outputs? (probably a GetOutputName() is broken and never returns NO_OUTPUT)');
      end;
      FOutputs.Clear;
      FOutCache := FOutputs.FSignals;
      while OutputCount < aCount
      do begin
        FOutputs.AddSignal( 0.0);
        FOutCache := FOutputs.FSignals;
      end;
    end;

    procedure   TMod.SetInputs( aStartIndex: Integer; const aValue: TSignals);
    begin
      FInputs .SetSignals( aStartIndex, aValue);
      FShadows.SetSignals( aStartIndex, aValue);
      FInCache     := FInputs .FSignals;
      FShadowCache := FShadows.FSignals;
    end;

    procedure   TMod.HandleParam( const aControl: string; aValue: TSignal); // virtual;
    // Handle a parameter set from the editor patch
    begin
      InternalValue[ aControl] := aValue;
    end;

//  public

    constructor TMod.Create( aParent: TSynthPatch; const aName: string); // virtual;
    begin
      inherited Create;
      FIsFast      := True;
      FIsMidi      := False;
      FTotalClocks := 0;
      FTotalZips   := 0;
      FParent      := aParent;
      Name         := aName;
      FInPuts      := TSignals.Create( 0);
      FShadows     := TSignals.Create( 0);
      FOutputs     := TSignals.Create( 0);
      FixOutputs;
      FixInputs;
      SetDefaults;
    end;

    destructor  TMod.Destroy; // override;
    begin
      FParent := nil;
      FreeAndNil( FOutputs);
      FreeAndNil( FInputs );
      FreeAndNil( FShadows);
      inherited;
    end;

    procedure   TMod.SetDefaults; // virtual;
    begin
    end;

    procedure   TMod.SaveKeyValue( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    var
      aBaseName : string;
      i         : Integer;
      Output    : string;
    begin
      aBaseName := '';
      for i := 1 to Length( aName)
      do begin
        if CharInSet( aName[ i], IdentifierStarters)
        then abaseName := abaseName + aName[ i]
        else Break;
      end;
      Output := Format( '%s', [ SignalToStr( aValue)], AppLocale);
      S := S + Format( Indent( anIndent + 1) + '%s %s' + sLineBreak, [ aName, Output], AppLocale);
    end;

    procedure   TMod.SaveInputValues( anIndent: Integer; var S: string);
    var
      i : Integer;
      N : string;
      p : Integer;
    begin
      p := 1;
      S := S + Indent( anIndent) + 'inputs(' + sLineBreak;
      for i := 0 to InputCount - 1
      do begin
        N := InputName( i);
        if SameText( N, NO_INPUT)
        then begin
          N := Format( 'in%d', [ p], AppLocale);
          Inc( p);
        end;
        SaveKeyValue( anIndent, N, FInCache[ i], S);
      end;
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TMod.SaveDetails( anIndent: Integer; var S: string); // virtual;
    begin
      S := S + Format( Indent( anIndent) + '%s' + sLineBreak, [ ClassName ], AppLocale);
      S := S + Format( Indent( anIndent) + '%s' + sLineBreak, [ Name      ], AppLocale);
      SaveInternals  ( anIndent, S);
      SaveInputValues( anIndent, S);
    end;

    procedure   TMod.SaveInternals( anIndent: Integer; var S: string); // virtual;
    var
      aName: string;
      i    : Integer;
    begin
      aName := InternalName( 0);
      if not SameText( aName, NO_INTERNAL)
      then begin
        S := S + Indent( anIndent) + 'internals(' + sLineBreak;
        i := 0;
        while not SameText( aName, NO_INTERNAL)
        do begin
          SaveKeyValue( anIndent, aName, GetInternal( aName), S);
          Inc( i);
          if i > 256
          then raise EInternalNotFound.Create( 'Are there really over 256 internals??');
          aName := InternalName( i);
        end;
        S := S + Indent( anIndent) + ')' + sLineBreak;
      end;
    end;

    procedure   TMod.SaveStrings( anIndent: Integer; var S: string); // virtual;
    var
      aName: string;
      i    : Integer;
    begin
      aName := StringName( 0);
      if not SameText( aName, NO_STRING)
      then begin
        S := S + Indent( anIndent) + 'strings(' + sLineBreak;
        i := 0;
        while not SameText( aName, NO_STRING)
        do begin
          S := S + Format( Indent( anIndent + 1) + '%s %s' + sLineBreak, [ aName, StringValue[ aName]], AppLocale);
          Inc( i);
          if i > 256
          then raise EInternalNotFound.Create( 'Are there really over 256 strings??');
          aName := InternalName( i);
        end;
        S := S + Indent( anIndent) + ')' + sLineBreak;
      end;
    end;

    procedure   TMod.SetInternal( const aName: string; aValue: TSignal); // virtual;
    begin
      // Internals is an option not present here.
    end;

    function    TMod.GetInternal( const aName: string): TSignal; // virtual;
    begin
      // Internals is an option not present here.
      Result := UNDEFINED;
    end;

    procedure   TMod.Save( anIndent: Integer; var S: string);
    begin
      S := S + Indent( anIndent) + 'module(' + sLineBreak;
      SaveDetails  ( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TMod.Reset; // virtual;
    begin
      FResetFlag := True;
    end;

    procedure   TMod.Clear; // virtual;
    begin
    end;

    function    TMod.FindInput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;
      for i := 0 to InputCount - 1
      do begin
        if SameText( aPinName, InputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TMod.FindOutput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;
      for i := 0 to OutputCount - 1
      do begin
        if SameText( aPinName, OutputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TMod.InputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INPUT;
    end;

    function    TMod.InternalName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INTERNAL;
    end;

    function    TMod.StringName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_STRING;
    end;

    function    TMod.OutputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_OUTPUT;
    end;

    procedure   TMod.PreTick;
    begin
      {$Q-}
      Inc( FSamples);
      FClocks := GetCpuClockCycleCount;
      {$Q+}
    end;

    procedure   TMod.PostTick;
    begin
      {$Q-}
      FClocks      := GetCpuClockCycleCount - FClocks;
      FTotalClocks := FTotalClocks + FClocks;
      {$Q+}
    end;

    procedure   TMod.PreZipper;
    begin
      {$Q-}
      FZips := GetCpuClockCycleCount;
      {$Q+}
    end;

    procedure   TMod.PostZipper;
    begin
      {$Q-}
      FZips      := GetCpuClockCycleCount - FZips;
      FTotalZips := FTotalZips + FZips;
      {$Q+}
    end;

    procedure   TMod.Tick;
    begin
      PreTick;
      DoTick;
      PostTick;
    end;

    Procedure   TMod.SlowTick;
    begin
      PreTick;
      DoSlowTick;
      PostTick;
    end;

    procedure   TMod.DeZipper;
    begin
      PreZipper;
      DoDeZipper;
      PostZipper;
    end;

    procedure   TMod.DoTick; // virtual;
    begin
      // Process inputs into new outputs - audio rate
    end;

    procedure   TMod.DoSlowTick; // virtual;
    begin
      // Process inputs into new outputs - control rate
    end;

    procedure   TMod.DoDeZipper;
    var
      i : Integer;
    begin
      for i := 0 to Length( FInCache) - 1
      do begin
        if i in FKnobMap
        then FInCache[ i] := ( FShadowCache[ i] + 511 * FInCache[ i]) / 512;
      end;
    end;

    procedure   TMod.FixZippers; // virtual;
    var
      i : Integer;
    begin
      for i := 0 to Length( FInCache) - 1
      do begin
        if i in FKnobMap
        then FInCache[ i] := FShadowCache[ i];
      end;
    end;

    procedure   TMod.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // virtual;
    begin
      // Collect info to be sent to module lights - this module has no lights
    end;

    procedure   TMod.GatherData( const aPrefix: string; const aCallback: TDataHandler); // virtual;
    begin
      // Collect numerical info to be sent to module viewers - this module has no viewers
    end;

    procedure   TMod.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // virtual;
    begin
      // Collect string info to be sent to module viewers - this module has no viewers
    end;

    procedure   TMod.AcceptParam( const aPath: string; aValue: TSignal);
    // Accept knob data sent from visual modules
    var
      aPrefix: string;
      aParam  : string;
      p      : Integer;
    begin
      if ( ClassType = TSynthPatch)
      then HandleParam( aPath, aValue)
      else begin
        ParsePrefix( aPath, aPrefix, aParam);
        if SameText( aPrefix, Name)                      // is the message for us?
        then begin
          if Pos( '/', aParam) = 0                       // When no more slashes left
          then begin                                     // we must be at the final bit ...
            p := Pos( '_', aParam);                      // The xxx_ is a conveniance in the "desiger" ... strip it here
            if p > 0
            then begin
              aParam := Copy( aParam, p + 1, Length( aParam));
              HandleParam( aParam, aValue);
            end;
          end;
        end;
      end;
    end;

    function    TMod.GetProfileInfoHtml: string;
    begin
      Result :=
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            Name,
            ClassName,
            Samples,
            TotalClocks,
            TotalClocks / Samples
          ],
          AppLocale
        );
    end;

    function    TMod.IsSameModule( const aModule: TMod): Boolean;
    begin
      Result := False;
      if Assigned( aModule)
      then Result := ClassType = aModule.ClassType;
    end;

    procedure   TMod.AcceptMidi( const aMsg: TMidiMessage); // virtual;
    begin
      // Nothing here, if is for chuldren, just building daydreams ...
    end;

    procedure   TMod.SendMidi( const aMsg: TMidiMessage);
    begin
      if Assigned( FParent)
      then FParent.SendMidi( aMsg);
    end;


{ ========
  TModInput = class( TMod)
  public
}

const

  input_out1 = 0;
  input_out2 = 1;

    procedure   TModInput.DoTick; // override;
    begin
    end;

    function    TModInput.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        input_out1 : Result := 'out1';
        input_out2 : Result := 'out2';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModOutput = class( TMod)
  public
}

const

  output_in1 = 0;
  output_in2 = 1;

    function    TModOutput.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        output_in1 : Result := 'in1';
        output_in2 : Result := 'in2';
        else         Result := NO_INPUT;
      end;
    end;


{ ========
  TSynthPatch = class
  private
    FVersion          : Integer;
    FModules          : TModuleArray;
    FConnections      : TConnections;
    FInputMappings    : TConnections;
    FOutputMappings   : TConnections;
    FDecimatorCount   : Integer;
    FInputQueue       : TSignalQueue;
    FWireClocks       : Int64;
    FTotalWireClocks  : Int64;
    FOnSendMidi       : TOnSendMidi;
  public
    property    Version                      : Integer     read FVersion       write FVersion;
    property    ModuleCount                  : Integer     read GetModuleCount;
    property    Module    [ anIndex: Integer]: TMod        read GetModule;
    property    ConnectionCount              : Integer     read GetConnectionCount;
    property    Overflows                    : Integer     read GetOverflows;
    property    Underflows                   : Integer     read GetUnderflows;
    property    OnSendMidi                   : TOnSendMidi read FOnSendMidi    write FOnSendMidi;
  private
}

    function    TSynthPatch.AssertUniqueName( const aName: string): string; // override;
    var
      i : Integer;
    begin
      // calling inherited here would tend to go recursive ...
      Result := aName;
      for i := 0 to ModuleCount - 1
      do begin
        if SameText( FModules[ i].Name, aName)
        then raise EModule.CreateFmt( 'module name "%s" is not unique.', [ aName]);
      end;
    end;

//  private

    function    TSynthPatch.CheckModuleIndex( anIndex: Integer): Boolean;
    begin
      Result := ( anIndex >= 0) and ( anIndex < ModuleCount);
    end;

    function    TSynthPatch.GetModuleCount: Integer;
    begin
      Result := Length( FModules);
    end;

    function    TSynthPatch.GetModule( anIndex: Integer): TMod;
    begin
      Result := nil;
      if CheckModuleIndex( anIndex)
      then Result := FModules[ anIndex];
    end;

    function    TSynthPatch.GetConnectionCount: Integer;
    begin
      Result := 0;
      if Assigned( FConnections)
      then Result := FConnections.Count;
    end;

    function    TSynthPatch.GetOverflows: Integer;
    begin
      if Assigned( FInputQueue)
      then Result := FInputQueue.Overflows
      else Result := 0;
    end;

    function    TSynthPatch.GetUnderflows: Integer;
    begin
      if Assigned( FInputQueue)
      then Result := FInputQueue.Underflows
      else Result := 0;
    end;

//  protected

    procedure   TSynthPatch.HandleParam( const aPath: string; aValue: TSignal); // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].AcceptParam( aPath, aValue);
    end;


//  public

    constructor TSynthPatch.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast      := True;
      FIsMidi      := True;
      FVersion     := 0;
      FInputQueue  := TSignalQueue.Create( 32768);
      FConnections := TConnections.Create( Self);
    end;

    destructor  TSynthPatch.Destroy; // override;
    begin
      FreeAndNil( FConnections   );  // For a faster clear
      Clear;                         // Clears the modules - that is it removes and frees them
      FreeAndNil( FInputQueue    );
      inherited;
    end;

    procedure   TSynthPatch.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].SetDefaults;
    end;

    procedure   TSynthPatch.SaveConnections( anIndent: Integer; var S: string);
    begin
      if Assigned( FConnections)
      then FConnections.Save( anIndent, 'connections', S);
    end;

    procedure   TSynthPatch.SaveModules( anIndent: Integer; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + 'modules(' + sLineBreak;
      for i := 0 to ModuleCount - 1
      do Module[ i].Save( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TSynthPatch.SaveDetails( anIndent: Integer; var S: string); // override;
    begin
      inherited;
      S := S + Indent( anIndent) + 'version ' + IntToStr( Version) + sLineBreak;
      SaveModules    ( anIndent, S);
      SaveConnections( anIndent, S);
    end;

    procedure   TSynthPatch.Clear;   // Clear moduleas and connections
    begin
      ClearModules;
      inherited Clear;
    end;

    procedure   TSynthPatch.Reset;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].Reset;
    end;

    procedure   TSynthPatch.RemoveConnectionsFor( aModuleIndex: Integer);
    begin
      if Assigned( FConnections)
      then FConnections.RemoveConnectionsFor( aModuleIndex);
    end;

    procedure   TSynthPatch.ClearModules;
    var
      aModule : TMod;
    begin
      while ModuleCount > 0
      do begin
        aModule :=  FModules[ ModuleCount - 1];
        RemoveConnectionsFor( ModuleCount - 1);
        SetLength(  FModules, ModuleCount - 1);
        aModule.Free;
      end;
    end;

    function    TSynthPatch.FindModule( const aModuleName: string): Integer;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;
      for i := 0 to ModuleCount - 1
      do begin
        if SameText( aModuleName, Module[ i].Name)
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TSynthPatch.ModuleName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModuleCount)
      then Result := Module[ anIndex].Name
      else Result := NO_MODULE;
    end;

    function    TSynthPatch.FindInputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_PMC;
      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindInput( aPinName);
    end;

    function    TSynthPatch.FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_PMC;
      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindOutput( aPinName);
    end;

    function    TSynthPatch.InputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_INPUT;
      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].InputName( aPinIndex);
    end;

    function    TSynthPatch.OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_OUTPUT;
      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].OutputName( aPinIndex);
    end;

    procedure   TSynthPatch.AddModule( const aModule: TMod); // overload;
    begin
      SetLength( FModules, ModuleCount + 1);
      FModules[ ModuleCount - 1] := aModule;
    end;

    procedure   TSynthPatch.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer); // overload;
    begin
      if Assigned( FConnections)
      then FConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin);
    end;

    procedure   TSynthPatch.AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string); // overload;
    var
      aSrcM: Integer;
      aSrcP: Integer;
      aDstM: Integer;
      aDstP: Integer;
    begin
      aSrcM := FindModule( aSrcModule);
      if aSrcM >= 0
      then begin
        aSrcP := FindOutputPin( aSrcM, aSrcPin);
        if aSrcP >= 0
        then begin
          aDstM := FindModule( aDstModule);
          if aDstM >= 0
          then begin
            aDstP := FindInputPin( aDstM, aDstPin);
            if aDstP >= 0
            then AddConnection( aSrcM, aSrcP, aDstM, aDstP)
            else raise EInputNotFound.CreateFmt( 'input "%s" not found on destination module "%s"', [ aDstPin, aDstModule]);
          end
          else raise EModuleNotFound.CreateFmt( 'destination module "%s" not found', [ aDstModule]);
        end
        else raise EOutputNotFound.CreateFmt( 'output "%s" not found on source module "%s"', [ aSrcPin, aSrcModule]);
      end
      else raise EModuleNotFound.CreateFmt( 'source module "%s" not found', [ aSrcModule]);
    end;

    procedure   TSynthPatch.PreWires;
    begin
      {$Q-}
      FWireClocks := GetCpuClockCycleCount;
      {$Q+}
    end;

    procedure   TSynthPatch.PostWires;
    begin
      FWireClocks      := GetCpuClockCycleCount - FWireClocks;
      FTotalWireClocks := FTotalWireClocks      + FWireClocks;
    end;

    procedure   TSynthPatch.PropagateSignals;
    // Copy signals from inputs to outputs
    begin
      PreWires;
      FConnections.PropagateSignals( FModules);
      PostWires;
    end;

const

  synthpatch_in1  = 0;
  synthpatch_in2  = 1;

  synthpatch_out1 = 0;
  synthpatch_out2 = 1;

    function    TSynthPatch.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        synthpatch_in1 : Result := 'in1';
        synthpatch_in2 : Result := 'in2';
        else             Result := NO_INPUT;
      end;
    end;

    function    TSynthPatch.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        synthpatch_out1 : Result := 'out1';
        synthpatch_out2 : Result := 'out2';
        else              Result := NO_OUTPUT;
      end;
    end;

    procedure   TSynthPatch.DoTick; // Override;
    var
      i         : Integer;
      p         : Integer;
      aModCount : Integer;
      aMod      : TMod;
      aPair     : TSignalPair;
      L         : TSignal;
      R         : TSignal;
    begin
      aModCount       := ModuleCount;
      aPair           := FInputQueue.GetData;           // Get patch external inputs
      p               := 0;
      L               := 0;
      R               := 0;
      FDecimatorCount := ( FDecimatorCount + 1) mod Slow_Decimation;

   // ApplyInputMappings;

      for i := 0 to aModCount - 1                    // Tick the modules in the patch
      do begin
        aMod := Module[ i];
        with aMod
        do begin

          if aMod is TModInput                // When it is an input module it needs the current in sample
          then begin
            FOutCache[ 0] := aPair.Left;
            FOutCache[ 1] := aPair.Right;
          end;


          if FIsFast
          then Tick;

          if p mod Slow_Decimation = FDecimatorCount
          then begin
            SlowTick;
            DeZipper;
          end;
          Inc( p);

          if aMod is TModOutput               // When the module is an output module add its inputs to our outputs
          then begin
            L := L + FInCache[ 0];
            R := R + FInCache[ 1];
          end;
        end;
      end;
      PropagateSignals;                       // Transfer module outputs to module inputs

      FOutCache[ 0] := L;                     // And fix our own outputs
      FOutCache[ 1] := R;
    end;

    procedure   TSynthPatch.FixZippers; // override;
    var
      i  : Integer;
    begin
      inherited;
      for i := 0 to ModuleCount - 1
      do Module[ i].FixZippers;
    end;

    procedure   TSynthPatch.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i  : Integer;
    begin
      if aPrefix = ''
      then begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherLights( Name, aCallback);
      end
      else begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherLights( aPrefix + '/' + Name, aCallback);
      end;
    end;

    procedure   TSynthPatch.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      i  : Integer;
    begin
      if aPrefix = ''
      then begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherData( Name, aCallback);
      end
      else begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherData( aPrefix + '/' + Name, aCallback);
      end;
    end;

    procedure   TSynthPatch.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    var
      i  : Integer;
    begin
      if aPrefix = ''
      then begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherInfo( Name, aCallback);
      end
      else begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherInfo( aPrefix + '/' + Name, aCallback);
      end;
    end;

    procedure   TSynthPatch.AcceptSamplePair( aLeft, aRight: TSignal);
    begin
      FInputQueue.Append( aLeft, aRight);
    end;

    function    TSynthPatch.GetProfileHtml: TStringList;
    var
      i          : Integer;
      SumClocks  : Int64;
      SumSamples : Int64;
      SumZips    : Int64;
    begin
      SumClocks  := 0;
      SumSamples := 0;
      SumZips    := 0;
      Result := TStringList.Create;
      Result.Add( '<!doctype html>');
      Result.Add( '<html><body><table><tbody>');
      Result.Add( '<tr><td>name</td><td>type</td><td>samples</td><td>clocks</td><td>clocks / sample</td></tr>');
      for i := 0 to ModuleCount - 1
      do begin
        Result.Add( FModules[ i].GetProfileInfoHtml);
        Inc( SumClocks , FModules[ i].TotalClocks);
        Inc( SumSamples, FModules[ i].Samples    );
        Inc( SumZips   , FModules[ i].TotalZips  );
      end;
      Result.Add( GetProfileInfoHtml);
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'total',
            'average module',
            SumSamples,
            SumClocks,
            SumClocks / SumSamples
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'dezippers',
            '&nbsp;',
            Samples,
            SumZips,
            SumZips / Samples
          ],
          AppLocale
        )
      );
      if ModuleCount > 0
      then begin
        Result.Add(
          Format(
            '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
            [
              'dezippers / module',
              '&nbsp;',
              Samples,
              Round( SumZips / ModuleCount),
              SumZips / Samples / ModuleCount
            ],
            AppLocale
          )
        );
      end;
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'wires',
            '&nbsp;',
            Samples,
            FTotalWireClocks,
            FTotalWireClocks / Samples
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%.0f</td><td align="right">%f</td></tr>',
          [
            'wire',
            'average',
            Samples,
            FTotalWireClocks / ConnectionCount,
            FTotalWireClocks / ConnectionCount / Samples
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'overhead',
            'lights etc.',
            Samples,
            TotalClocks - SumClocks - FTotalWireClocks - SumZips,
            ( TotalClocks - SumClocks - FTotalWireClocks - SumZips) / Samples
          ],
          AppLocale
        )
      );
      Result.Add( '<tr><td>name</td><td>type</td><td>samples</td><td>clocks</td><td>clocks / sample</td></tr>');
      Result.Add( '</tbody></table></body></html>');
    end;

    function    TSynthPatch.IsSamePatch( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then begin
        Result :=
          SameModules    ( aPatch) and
          SameConnections( aPatch)
      end;
    end;

    function    TSynthPatch.SameModules( const aPatch: TSynthPatch): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      if Assigned( aPatch)
      then begin
        if ModuleCount = aPatch.ModuleCount
        then begin
          Result := True;
          for i := 0 to ModuleCount - 1
          do begin
            if not Module[ i].IsSameModule( aPatch.Module[ i])
            then begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;

    function    TSynthPatch.SameConnections( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then Result := FConnections.IsSameConnections( aPatch.FConnections);
    end;

    procedure   TSynthPatch.AcceptMidi( const aMsg: TMidiMessage); // overide;
    var
      i : Integer;
    begin
      if FIsMidi
      then begin
        for i := 0 to ModuleCount - 1
        do begin
          if Module[ i].FIsMidi
          then Module[ i].AcceptMidi( aMsg);
        end;
      end;
    end;

    procedure   TSynthPatch.SendMidi( const aMsg: TMidiMessage); // override;
    begin
      if Assigned( FOnSendMidi)
      then FOnSendMidi( Self, aMsg);
    end;


end.
