unit module_implementations;

{

   COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Math, WaveUtils, WaveStorage,

  Globals, module_defs, module_matrix, KnobsUtils, KnobsConversions;

type

  TARState      = ( arIdle, arA, arR, arWait);
  TAHDState     = ( ahdIdle, ahdA, ahdH, ahdD, ahdWait);
  TADSRState    = ( adsrIdle, adsrA, adsrD, adsrS, adsrR, adsrWait);
  TEnvRange     = ( erSlow, erMedium, erFast);
  TEnvShape     = ( esLog, esLin, esExp);
  TLfoRange     = ( lsSlow, lsMedium, lsFast, lsBPM, lsLin);
  TDelaySize    = ( dsShort, dsMedium, dsLong);
  TRWState      = ( trwIdle, trwFired);
  TTrigOScState = ( toIdle, toRunning, toWaiting, toDone);
  TSeqInState   = ( seqIdle, seqHigh);
  TSeqStepState = ( sssIdle, sssActive, sssWaiting);

  TModConstant = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModInverter = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModAdder = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMultiplier = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModGate = class( TMod)
  private
    FFunction : Integer;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDivider = class( TMod)
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FState     : Boolean;
  public
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDFlipFlop = class( TMod)
  private
    FClockState : Boolean;
  public
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModRSFlipFlop = class( TMod)
  private
    FSetState   : Boolean;
    FResetState : Boolean;
  public
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNot = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModAmplifier = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    procedure   SetDefaults;                                                                     override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModScaler = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    procedure   SetDefaults;                                                                     override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRatio = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    procedure   SetDefaults;                                                                     override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRotator = class( TMod)
  private
    FAngle: TSignal;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRectifier = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMinMax = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModIntDif = class( TMod)
  private
    FSum        : TSignal;
    FlastSample : TSignal;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModBaseOsc = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldSync  : Boolean;
    FShape    : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModBaseTrigOsc = class( TModBaseOsc)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
  public
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
  end;

  TModOsc = class( TModBaseOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModTrigOsc = class( TModBaseTrigOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModSquare = class( TModOsc)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareTrig = class( TModTrigOsc)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModBaseLfo = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldRes   : Boolean;
    FRange    : Integer;
    FShape    : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModBaseTrigLfo = class( TModBaseLfo)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
  public
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
  end;

  TModLfo = class( TModBaseLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModLfoTrig = class( TModBaseTrigLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModSquareLfo = class( TModBaseLfo)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareLfoTrig = class( TModBaseTrigLfo)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModRandomWalkLfo = class( TModBaseLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoise = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    procedure   SetDefaults;                                                                     override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseTrig = class( TMod)
  private
    FLastTrig: Boolean;
  public
    procedure   DoTick;                                                                          override;
    procedure   SetDefaults;                                                                     override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfo = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfoTrig = class( TMod)
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAverage = class( TMod)
  private
    FAlpha : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSVF = class( TMod)
  private
    FFreq  : TSignal;
    FDamp  : TSignal;
    FDrive : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAHD = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TAHDState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvADSR = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TADSRState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps   : Integer;
    FCounter : Integer;
    FInState : TSeqInState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModTextSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps    : Integer;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    procedure   SetStepValues( const aValues: TSignals);
  end;

  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMemory = class( TMod)
  private
    FMemoryL : TSignals;
    FMemoryR : TSignals;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    function    Convolve( aSample: TSignal): TSignalPair;
  end;

  TModDelay = class( TModMemory)
  private
    FMemLength : TSignal;
    FInPointer : Integer;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModWave = class( TModMemory)
  private
    FFileName    : string;
    FDuration    : string;
    FIsStereo    : Boolean;
  private
    procedure   SetDuration( aValue: Cardinal);
    procedure   SwapMemoryFor( aMemory: TSignalArray);
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
  public
    destructor  Destroy;                                                                         override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    StringName( anIndex: Integer): string;                                           override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTrackAndHold = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSwitch2to1 = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModXFade = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPan = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM4to1 = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixS2to1 = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM6to6 = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModQuantize = class( TMod)
  public
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDisplay = class( TMod)
  private
    FMax1: TSignal;
    FMax2: TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModStateSpace = class( TMod)
  private
    FA        : TSignalMatrix;
    FB        : TSignalMatrix;
    FC        : TSignalMatrix;
    FD        : TSignalMatrix;
    FX        : TSignalMatrix;
    FNewX     : TSignalMatrix;
    FOut      : TSignalMatrix;
    FIn       : TSignalMatrix;
    FRunnable : Boolean;
  private
    function    Consistent: Boolean;
    procedure   ClearState;
  protected
    procedure   Setup;                                                                  virtual; abstract;
    procedure   Run;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMidiIn = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModMidiNoteIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FLastNote    : Byte;
    FLastChannel : Byte;
    FOldChannel  : Byte;
    FGate        : Boolean;
    FNote        : TSignal;
    FVelocityOn  : TSignal;
    FVelocityOff : TSignal;
  public
    procedure   NotesOff;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMidiCCIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FCC          : Byte;
    FGate        : Boolean;
    FValue       : TSignal;
    FCounter     : Integer;
    FLedCounter  : Integer;
  public
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMidiOut = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModMidiNoteOut = class( TModMidiOut)
  private
    FLastNote    : Byte;
    FLastChannel : Byte;
    FGate        : Boolean;
    FDidSend     : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModMidiCCOut = class( TModMidiOut)
  private
    FGate  : Boolean;
  public
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;



  function SignalToEnvRange ( aValue: TSignal): TEnvRange;
  function SignalToEnvShape ( aValue: TSignal): TEnvShape;
  function SignalToLfoRange ( aValue: TSignal): TLfoRange;
  function SignalToDelaySize( aValue: TSignal): TDelaySize;



implementation



  function SignalToEnvRange( aValue: TSignal): TEnvRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := erFast
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := erMedium
    else Result := erSlow;
  end;

  function SignalToEnvShape( aValue: TSignal): TEnvShape;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := esLog
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := esLin
    else Result := esExp;
  end;

  function SignalToLfoRange( aValue: TSignal): TLfoRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := lsFast
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := lsMedium
    else if AlmostEqual( aValue, 2, 0.45)
    then Result := lsSlow
    else if AlmostEqual( aValue, 3, 0.45)
    then Result := lsBPM
    else Result := lsLin;
  end;

  function SignalToDelaySize( aValue: TSignal): TDelaySize;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := dsShort
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := dsMedium
    else Result := dsLong;
  end;


{ ========
  TModConstant = class( TMod)
  public
}

const

  constant_value = 0;
  constant_mute  = 1;

  constant_out   = 0;

    constructor TModConstant.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModConstant.SetDefaults; // override;
    begin
      FKnobMap := [ constant_value, constant_mute];
    end;

    procedure   TModConstant.DoSlowTick; // override;
    begin
      FOutCache[ constant_out] := FInCache[ constant_value] * SignalToMute( FInCache[ constant_mute]);
    end;

    function    TModConstant.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        constant_value :  Result := 'value';
        constant_mute  :  Result := 'mute';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModConstant.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        constant_out :  Result := 'out';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModInverter = class( TMod)
  public
}

const

  inverter_in1  = 0; // Keep these consecutive
  inverter_in2  = 1;
  inverter_in3  = 2;
  inverter_in4  = 3;

  inverter_out1 = 0; // Keep these consecutive
  inverter_out2 = 1;
  inverter_out3 = 2;
  inverter_out4 = 3;

    procedure   TModInverter.DoTick; // override;
    var
      i : Integer;
    begin
      for i := inverter_in1 to inverter_in4
      do FOutCache[ i - inverter_in1 + inverter_out1] := - FInCache[ i];
    end;

    function    TModInverter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        inverter_in1 : Result := 'in1';
        inverter_in2 : Result := 'in2';
        inverter_in3 : Result := 'in3';
        inverter_in4 : Result := 'in4';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModInverter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        inverter_out1 : Result := 'out1';
        inverter_out2 : Result := 'out2';
        inverter_out3 : Result := 'out3';
        inverter_out4 : Result := 'out4';
        else            Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModAdder = class( TMod)
  public
}

const

  adder_mute = 0;
  adder_in1  = 1;
  adder_in2  = 2;
  adder_in3  = 3;
  adder_in4  = 4;

  adder_out  = 0;

    procedure   TModAdder.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := adder_in1 to adder_in4          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInCache[ i] := 0;
      FKnobMap := [ adder_mute];
    end;

    procedure   TModAdder.DoTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := 0.0;
      for i := adder_in1 to adder_in4
      do R := R + FInCache[ i];
      FOutCache[ adder_out] := R * SignalToMute( FInCache[ adder_mute]);
    end;

    function    TModAdder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        adder_mute : Result := 'mute';
        adder_in1  : Result := 'in1';
        adder_in2  : Result := 'in2';
        adder_in3  : Result := 'in3';
        adder_in4  : Result := 'in4';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModAdder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        adder_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModMultiplier = class( TMod)
  public
}

const

  multiplier_mute = 0;
  multiplier_in1  = 1;
  multiplier_in2  = 2;
  multiplier_in3  = 3;
  multiplier_in4  = 4;

  multiplier_out  = 0;

    procedure   TModMultiplier.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := multiplier_in1 to multiplier_in4          // Make sure unused inputs are set to 1, so it'll 'work'
      do FInCache[ i] := 1;
      FKnobMap := [ multiplier_mute];
    end;

    procedure   TModMultiplier.DoTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := 1.0;
      for i := multiplier_in1 to multiplier_in4
      do R := R * FInCache[ i];
      FOutCache[ multiplier_out] := AntiRunAway( R * SignalToMute( FInCache[ multiplier_mute]));
    end;

    function    TModMultiplier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multiplier_mute : Result := 'mute';
        multiplier_in1  : Result := 'in1';
        multiplier_in2  : Result := 'in2';
        multiplier_in3  : Result := 'in3';
        multiplier_in4  : Result := 'in4';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModMultiplier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multiplier_out : Result := 'out';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModLogic = class( TMod)
  public
}

const

  gate_function = 0;
  gate_in1      = 1;
  gate_in2      = 2;
  gate_in3      = 3;
  gate_in4      = 4;

  gate_out      = 0;

    procedure   TModGate.SetDefaults; // override;
    var
      i : Integer;
    begin
      case SignalToInt( FInCache[ gate_function]) of
        0  : // and
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 1, so it'll 'work'
            do FInCache[ i] := 1;
          end;
        1  : // nand
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 1, so it'll 'work'
            do FInCache[ i] := 1;
          end;
        2  : // or
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInCache[ i] := 0;
          end;
        3  : // nor
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInCache[ i] := 0;
          end;
        4  : // xor
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInCache[ i] := 0;
          end;
        else // xnor
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInCache[ i] := 0;
          end;
      end;
    end;

    procedure   TModGate.DoTick; // override;
    var
      i         : Integer;
      R         : Boolean;
      OneCounts : Integer;
      aFunction : Integer;
    begin
      aFunction := SignalToInt( FInCache[ gate_function]);
      if aFunction <> FFunction
      then begin
        FFunction := aFunction;
        SetDefaults;
      end;

      case aFunction of
        0  : // and
          begin
            R := True;
            for i := gate_in1 to gate_in4
            do R := R and SignalToLogic( FInCache[ i]);
            FOutCache[ gate_out] := LogicToSignal( R);
          end;
        1  : // nand
          begin
            R := True;
            for i := gate_in1 to gate_in4
            do R := R and SignalToLogic( FInCache[ i]);
            FOutCache[ gate_out] := LogicToSignal( not R);
          end;
        2  : // or
          begin
            R := False;
            for i := gate_in1 to gate_in4
            do R := R or SignalToLogic( FInCache[ i]);
            FOutCache[ gate_out] := LogicToSignal( R);
          end;
        3  : // nor
          begin
            R := False;
            for i := gate_in1 to gate_in4
            do R := R or SignalToLogic( FInCache[ i]);
            FOutCache[ gate_out] := LogicToSignal( not R);
          end;
        4  : // xor
          begin
            OneCounts := 0;
            for i := gate_in1 to gate_in4
            do begin
              if SignalToLogic( FInCache[ i])
              then Inc( OneCounts);
            end;
            FOutCache[ gate_out] := LogicToSignal( OneCounts = 1);
          end;
        else // xnor
          begin
            OneCounts := 0;
            for i := gate_in1 to gate_in4
            do begin
              if SignalToLogic( FInCache[ i])
              then Inc( OneCounts);
            end;
            FOutCache[ gate_out] := LogicToSignal( OneCounts <> 1);
          end;
      end;
    end;

    procedure   TModGate.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutCache[ gate_out]], AppLocale));
    end;

    function    TModGate.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gate_function : Result := 'function';
        gate_in1      : Result := 'in1';
        gate_in2      : Result := 'in2';
        gate_in3      : Result := 'in3';
        gate_in4      : Result := 'in4';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModGate.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gate_out : Result := 'out';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDivider = class( TMod)
  private
    FLastInput: Boolean;
    FCounter  : Integer;
    FState    : Boolean;
  public
}

const

  divider_count = 0;
  divider_trig  = 1;

  divider_out   = 0;

    procedure   TModDivider.DoTick; // override;
    const
      States : array[ Boolean] of TSignal = ( -1, 1);
    var
      Inp: Boolean;
    begin
      if ResetFlag
      then begin
        FCounter   := Max( 0, Round( FInCache[ divider_count] - 1));
        FState     := False;
        FOutCache[ divider_out] := States[ FState];
      end
      else begin
        Inp := SignalToLogic( FInCache[ divider_trig]);
        if FLastInput xor Inp
        then begin
          if FCounter <= 0
          then begin
            FCounter   := Max( 0, Round( FInCache[ divider_count] - 1));
            FState     := not FState;
            FOutCache[ divider_out] := States[ FState];
          end
          else Dec( FCounter);
        end;
        FLastInput := Inp;
      end;
    end;

    procedure   TModDivider.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutCache[ divider_out]], AppLocale));
    end;

    function    TModDivider.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        divider_count : Result := 'count';
        divider_trig  : Result := 'trig';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModDivider.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        divider_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDFlipFlop = class( TMod)
  private
    FClockState : Boolean;
  public
}

const

  dflipflop_d    = 0;
  dflipflop_clk  = 1;

  dflipflop_q    = 0;
  dflipflop_notq = 1;


    procedure   TModDFlipFlop.DoTick; // override;
    var
      CLK : Boolean;
      Q   : Boolean;
    begin
      if ResetFlag
      then begin
        FOutCache[ dflipflop_q] := LogicToSignal( False);
        Q                       := False;
        ResetFlag               := False;
      end
      else begin
        CLK := SignalToLogic( FInCache[ dflipflop_clk]);
        if not FClockState and CLK
        then begin
          Q                       := SignalToLogic( FInCache[ dflipflop_d]);
          FOutCache[ dflipflop_q] := LogicToSignal( Q);
        end
        else Q := SignalToLogic( FOutCache[ dflipflop_q]);
        FClockState := CLK;
      end;
      FOutCache[ dflipflop_notq] := LogicToSignal( not Q);
    end;

    procedure   TModDFlipFlop.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/q/%f'   , [ aPrefix, Name, FOutCache[ dflipflop_q   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/notq/%f', [ aPrefix, Name, FOutCache[ dflipflop_notq]], AppLocale));
    end;

    function    TModDFlipFlop.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        dflipflop_d   : Result := 'd';
        dflipflop_clk : Result := 'clk';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModDFlipFlop.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        dflipflop_q    : Result := 'q';
        dflipflop_notq : Result := 'notq';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRSFlipFlop = class( TMod)
  private
    FSetState   : Boolean;
    FResetState : Boolean;
  public
}

const

  rsflipflop_s    = 0;
  rsflipflop_r    = 1;

  rsflipflop_q    = 0;
  rsflipflop_notq = 1;

    procedure   TModRSFlipFlop.DoTick; // override;
    var
      S : Boolean;
      R : Boolean;
      Q : Boolean;
    begin
      S := SignalToLogic( FInCache [ rsflipflop_s]);
      R := SignalToLogic( FInCache [ rsflipflop_r]) or ResetFlag;
      Q := SignalToLogic( FOutCache[ rsflipflop_q]);
      if not FSetState and S
      then Q := True;
      if not FResetState and R
      then Q := False;
      FOutCache[ rsflipflop_q   ] := LogicToSignal(     Q);
      FOutCache[ rsflipflop_notq] := LogicToSignal( not Q);
      FSetState   := S;
      FResetState := R;
      ResetFlag   := False;
    end;

    procedure   TModRSFlipFlop.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/q/%f'   , [ aPrefix, Name, FOutCache[ rsflipflop_q   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/notq/%f', [ aPrefix, Name, FOutCache[ rsflipflop_notq]], AppLocale));
    end;

    function    TModRSFlipFlop.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rsflipflop_s : Result := 's';
        rsflipflop_r : Result := 'r';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModRSFlipFlop.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rsflipflop_q    : Result := 'q';
        rsflipflop_notq : Result := 'notq';
        else              Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNot = class( TMod)
  public
}

const

  not_in1  = 0;
  not_in2  = 1;
  not_in3  = 2;
  not_in4  = 3;

  not_out1 = 0;
  not_out2 = 1;
  not_out3 = 2;
  not_out4 = 3;

    procedure   TModNot.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := not_in1 to not_in4          // Set unused inputs to 1 - just to keep the lights off
      do FInCache[ i] := 1;
    end;

    procedure   TModNot.DoTick; // override;
    var
      i : Integer;
    begin
      for i := not_in1 to not_in4
      do FOutCache[ i] := LogicToSignal( not SignalToLogic( FInCache[ i]));
    end;

    procedure   TModNot.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := not_out1 to not_out4
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, i + 1, FOutCache[ i]], AppLocale));
    end;

    function    TModNot.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        not_in1 : Result := 'in1';
        not_in2 : Result := 'in2';
        not_in3 : Result := 'in3';
        not_in4 : Result := 'in4';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModNot.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        not_out1 : Result := 'out1';
        not_out2 : Result := 'out2';
        not_out3 : Result := 'out3';
        not_out4 : Result := 'out4';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAmplifier = class( TMod)
  public
}

const

  amplifier_gain = 0;
  amplifier_mute = 1;
  amplifier_in1  = 2; // keep in1 .. in4 consecutive
  amplifier_in2  = 3;
  amplifier_in3  = 4;
  amplifier_in4  = 5;

  amplifier_out1 = 0; // keep out1 .. out4 consecutive
  amplifier_out2 = 1;
  amplifier_out3 = 2;
  amplifier_out4 = 3;

    procedure   TModAmplifier.DoTick; // override;
    var
      i : Integer;
      G : TSignal;
    begin
      G := AntiRunAway( FInCache[ amplifier_gain] * SignalToMute( FInCache[ amplifier_mute]));
      for i := amplifier_in1 to amplifier_in4
      do FOutCache[ i - amplifier_in1 + amplifier_out1] := AntiRunAway( FInCache[ i] * G);
    end;

    procedure   TModAmplifier.SetDefaults; // override;
    begin
      FKnobMap := [ amplifier_gain, amplifier_mute];
    end;

    function    TModAmplifier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        amplifier_gain : Result := 'gain';
        amplifier_mute : Result := 'mute';
        amplifier_in1  : Result := 'in1';
        amplifier_in2  : Result := 'in2';
        amplifier_in3  : Result := 'in3';
        amplifier_in4  : Result := 'in4';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModAmplifier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        amplifier_out1: Result := 'out1';
        amplifier_out2: Result := 'out2';
        amplifier_out3: Result := 'out3';
        amplifier_out4: Result := 'out4';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModScaler = class( TMod)
  public
}

const

  scaler_gain   = 0;
  scaler_offset = 1;
  scaler_mode   = 2;
  scaler_mute   = 3;
  scaler_in1    = 4;
  scaler_in2    = 5;
  scaler_in3    = 6;
  scaler_in4    = 7;

  scaler_out1   = 0;
  scaler_out2   = 1;
  scaler_out3   = 2;
  scaler_out4   = 3;


    procedure   TModScaler.DoTick; // override;
    var
      i : Integer;
      G : TSignal;
      O : TSignal;
    begin
      G := FInCache[ scaler_gain  ] * SignalToMute( FInCache[ scaler_mute]);
      O := FInCache[ scaler_offset] * SignalToMute( FInCache[ scaler_mute]);
      for i := scaler_in1 to scaler_in4
      do FOutCache[ i - scaler_in1 + scaler_out1] := AntiRunAway( G * FInCache[ i] + O);
    end;

    procedure   TModScaler.SetDefaults; // override;
    begin
      FKnobMap := [ scaler_gain, scaler_offset, scaler_mute];
    end;

    function    TModScaler.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        scaler_gain   : Result := 'gain';
        scaler_offset : Result := 'offset';
        scaler_mode   : Result := 'mode';
        scaler_mute   : Result := 'mute';
        scaler_in1    : Result := 'in1';
        scaler_in2    : Result := 'in2';
        scaler_in3    : Result := 'in3';
        scaler_in4    : Result := 'in4';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModScaler.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        scaler_out1 : Result := 'out1';
        scaler_out2 : Result := 'out2';
        scaler_out3 : Result := 'out3';
        scaler_out4 : Result := 'out4';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRatio = class( TMod)
  public
}

const

  ratio_multiplier = 0;
  ratio_divider    = 1;
  ratio_mute       = 2;
  ratio_in1        = 3;
  ratio_in2        = 4;
  ratio_in3        = 5;
  ratio_in4        = 6;

  ratio_out1       = 0;
  ratio_out2       = 1;
  ratio_out3       = 2;
  ratio_out4       = 3;


    procedure   TModRatio.DoTick; // override;
    var
      i      : Integer;
      MulDiv : TSignal;
    begin
      Muldiv := Round( SignalToMute( FInCache[ ratio_mute]) * FInCache[ ratio_multiplier]) / Round( FInCache[ ratio_divider]);
      for i := ratio_in1 to ratio_in4
      do FOutCache[ i - ratio_in1 + ratio_out1] := AntiRunAway( MulDiv * FInCache[ i]);
    end;

    procedure   TModRatio.SetDefaults; // override;
    begin
      FKnobMap := [ ratio_multiplier, ratio_divider, ratio_mute];
    end;

    function    TModRatio.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ratio_multiplier : Result := 'multiplier';
        ratio_divider    : Result := 'divider';
        ratio_mute       : Result := 'mute';
        ratio_in1        : Result := 'in1';
        ratio_in2        : Result := 'in2';
        ratio_in3        : Result := 'in3';
        ratio_in4        : Result := 'in4';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModRatio.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ratio_out1 : Result := 'out1';
        ratio_out2 : Result := 'out2';
        ratio_out3 : Result := 'out3';
        ratio_out4 : Result := 'out4';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRotator = class( TMod)
  public
}

const

  rotator_angle = 0;
  rotator_in1   = 1;
  rotator_in2   = 2;

  rotator_out1  = 0;
  rotator_out2  = 1;


    procedure   TModRotator.SetDefaults; // override;
    begin
      FInCache[ rotator_angle] := 0;
      FInCache[ rotator_in1  ] := 0;
      FInCache[ rotator_in2  ] := 0;
    end;

    procedure   TModRotator.DoTick; // override;
    var
      aCos    : TSignal;
      aSin    : TSignal;
      in1     : TSignal;
      in2     : TSignal;
    begin
      FAngle := FModulo( FInCache[ rotator_angle] + 64, 2) * PI;
      SinCos( FAngle, aSin, aCos);
      in1 := FInCache[ rotator_in1];
      in2 := FInCache[ rotator_in2];
      FOutCache[ rotator_out1] := aCos * in1 - aSin * in2;
      FOutCache[ rotator_out2] := aSin * in1 + aCos * in2;
    end;

    function    TModRotator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rotator_angle : Result := 'angle';
        rotator_in1   : Result := 'in1';
        rotator_in2   : Result := 'in2';
        else Result := NO_INPUT;
      end;
    end;

    function    TModRotator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rotator_out1 : Result := 'out1';
        rotator_out2 : Result := 'out2';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRectifier = class( TMod)
  public
}

const

  rectifier_mode   = 0;
  rectifier_in     = 1;

  rectifier_out    = 0;
  rectifier_outinv = 1;


    procedure   TModRectifier.SetDefaults; // override;
    begin
      FInCache[ rectifier_in] := 0;
    end;

    procedure   TModRectifier.DoTick; // override;
    var
      Mode : Integer;
      Inp  : TSignal;
    begin
      Mode := SignalToInt( FInCache[ rectifier_mode]);
      Inp  :=              FInCache[ rectifier_in  ] ;

      if Mode = 0                                      // Half wave
      then begin
        if SignalToLogic( Inp)
        then FOutCache[ rectifier_out] := Inp
        else FOutCache[ rectifier_out] := 0;
      end
      else if Mode = 1                                 // Full wave
      then FOutCache[ rectifier_out] := Abs( Inp)
      else FOutCache[ rectifier_out] :=      Inp;      // Off

      FOutCache[ rectifier_outinv] := - FOutCache[ rectifier_out];
    end;

    function    TModRectifier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rectifier_mode : Result := 'mode';
        rectifier_in   : Result := 'in';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModRectifier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rectifier_out    : Result := 'out';
        rectifier_outinv : Result := 'outinv';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMinMax = class( TMod)
  public
}

const

  minmax_in1 = 0;
  minmax_in2 = 1;

  minmax_min = 0;
  minmax_max = 1;


    procedure   TModMinMax.SetDefaults; // override;
    begin
      FInCache[ minmax_in1] := 0;
      FInCache[ minmax_in2] := 0;
    end;

    procedure   TModMinMax.DoTick; // override;
    var
      in1 : TSignal;
      in2 : TSignal;
    begin
      in1                    := FInCache[ minmax_in1];
      in2                    := FInCache[ minmax_in2];
      FOutCache[ minmax_min] := Min( in1, in2);
      FOutCache[ minmax_max] := Max( in1, in2);
    end;

    function    TModMinMax.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        minmax_in1 : Result := 'in1';
        minmax_in2 : Result := 'in2';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModMinMax.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        minmax_min : Result := 'min';
        minmax_max : Result := 'max';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModIntDif = class( TMod)
  private
    FSum        : TSignal;
    FlastSample : TSignal;
  public
}

const

  intdif_in  = 0;

  intdif_int = 0;
  intdif_dif = 1;


    procedure   TModIntDif.SetDefaults; // override;
    begin
      FInCache[ intdif_in] := 0;
    end;

    procedure   TModIntDif.DoTick; // override;
    var
      Inp : TSignal;
    begin
      Inp := AntiRunAway( FInCache[ intdif_in]);

      if ResetFlag
      then begin
        FSum        := 0;
        FlastSample := Inp;
        ResetFlag   := False;
      end
      else FSum := Clip( FSum + Inp, -1000, 1000);

      FOutCache[ intdif_int] := FSum;
      FOutCache[ intdif_dif] := AntiRunAway( Inp - FlastSample);
      FlastSample            := Inp;
    end;

    function    TModIntDif.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        intdif_in : Result := 'in';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModIntDif.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        intdif_int : Result := 'int';
        intdif_dif : Result := 'dif';
        else         Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModBaseOsc = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldSync  : Boolean;
  public
}

const

  baseosc_frequency = 0;
  baseosc_fm        = 1;
  baseosc_phase     = 2;
  baseosc_sync      = 3;
  baseosc_mute      = 4;
  baseosc_shape     = 5;
  baseosc_pmlevel   = 6;
  baseosc_fmlevel   = 7;
  baseosc_cents     = 8;

  baseosc_out       = 0;


    constructor TModBaseOsc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
    end;

    procedure   TModBaseOsc.SetDefaults; // override;
    begin
      FKnobMap := [
        baseosc_frequency,
        baseosc_pmlevel,
        baseosc_fmlevel,
        baseosc_cents
      ];
    end;

    procedure   TModBaseOsc.DoTick; // override;
    var
      Delta : TSignal;
    begin
      if ( not FOldSync) and SignalToLogic( FInCache[ baseosc_sync])
      then FPhase  := 0.0;
      FOldSync  := SignalToLogic( FInCache[ baseosc_sync]);
      Delta     := LookupPhaseDelta( FInCache[ baseosc_frequency] + FInCache[ baseosc_fm] * FInCache[ baseosc_fmlevel] + FInCache[ baseosc_cents]);
      FPhase    := FModulo( FPhase + Delta, 1.0);
      FPosition := FModulo( FPhase + Clip( FInCache[ baseosc_phase] * FInCache[ baseosc_pmlevel], -1, 1), 1.0);
      FShape    := SignalToInt( FInCache[ baseosc_shape]);
    end;

    function    TModBaseOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baseosc_frequency : Result := 'frequency';
        baseosc_fm        : Result := 'fm';
        baseosc_phase     : Result := 'phase';
        baseosc_sync      : Result := 'sync';
        baseosc_mute      : Result := 'mute';
        baseosc_shape     : Result := 'shape';
        baseosc_pmlevel   : Result := 'pmlevel';
        baseosc_fmlevel   : Result := 'fmlevel';
        baseosc_cents     : Result := 'cents';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModBaseOsc.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baseosc_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseTrigOsc = class( TModBaseOsc)
  private
    FState   : TTrigOscState;
    FCounter : TSignal;
  public
}

const

  basetrigosc_trig  = baseosc_phase;
  basetrigosc_count = baseosc_sync;


    procedure   TModbaseTrigOsc.DoTick; // override;
    var
      FOldCounter : Integer;
      Delta       : TSignal;
    begin

      case FState of

        toIdle :
          begin
            FPosition := 0.0;
            FCounter  := 0;
            if SignalToLogic( FInCache[ basetrigosc_trig])
            then FState := toRunning;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            Delta     := LookupPhaseDelta( FInCache[ baseosc_frequency] + FInCache[ baseosc_fm] * FInCache[ baseosc_fmlevel] + FInCache[ baseosc_cents]);
            FPosition := FModulo( FPosition + Delta, 1.0);
            FShape    := SignalToInt( FInCache[ baseosc_shape]);

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPosition, 0.5, Delta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPosition, 1.0, Delta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInCache[ basetrigosc_count]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPosition, 0.5, Delta) and
                not AlmostEqual( FPosition, 1.0, Delta)
              then FState := toRunning;
            end;

          end;

        toDone :
          begin
            FPosition := 0.0;
            if FInCache[ basetrigosc_trig] <= 0.0
            then FState := toIdle;
          end;
      end;
    end;

    function    TModBaseTrigOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        basetrigosc_trig  : Result := 'trig';
        basetrigosc_count : Result := 'count';
        else                Result := inherited;
      end;
    end;

    procedure   TModBaseTrigOsc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ basetrigosc_trig]], AppLocale));
    end;


{ ========
  TModOsc = class( TModBaseOsc)
  public
}

    procedure   TModOSc.DoTick; // override;
    begin
      inherited;
      if SignalToLogic( FInCache[ baseosc_mute])
      then FOutCache[ baseosc_out] := 0
      else begin
        case FShape of
          0  : // sine
            FOutCache[ baseosc_out] := LookupSine( 2 * FPosition);
          1  : // tri
            if FPosition <= 0.5
            then FOutCache[ baseosc_out] := AntiRunAway( 4.0 *         FPosition  - 1)
            else FOutCache[ baseosc_out] := AntiRunAway( 4.0 * ( 1.0 - FPosition) - 1);
          2  : // saw
            FOutCache[ baseosc_out] := AntiRunAway( 2 * FPosition - 1);
          3  : // square
            if FPosition < 0.5
            then FOutCache[ baseosc_out] := -1.0
            else FOutCache[ baseosc_out] :=  1.0;
        end;
      end;
    end;


{ ========
  TModTrigOsc = class( TModBaseTrigOsc)
  public
}

    procedure   TModTrigOsc.DoTick; // override;
    begin
      inherited;
      if SignalToLogic( FInCache[ baseosc_mute])
      then FOutCache[ baseosc_out] := 0
      else begin
        case FShape of
          0  : // sine
            FOutCache[ baseosc_out] := LookupSine( 2 * FPosition);
          1  : // tri
            if FPosition <= 0.5
            then FOutCache[ baseosc_out] := AntiRunAway( 4.0 *         FPosition  - 1)
            else FOutCache[ baseosc_out] := AntiRunAway( 4.0 * ( 1.0 - FPosition) - 1);
          2  : // saw
            FOutCache[ baseosc_out] := AntiRunAway( 2 * FPosition - 1);
          3  : // square
            if FPosition < 0.5
            then FOutCache[ baseosc_out] := -1.0
            else FOutCache[ baseosc_out] :=  1.0;
        end;
      end;
    end;



{ ========
  TModSquare = class( TModOsc)
  public
}

const

  square_pwm      =  9;
  square_pwmlevel = 10;


    procedure   TModSquare.SetDefaults;
    begin
      inherited;
      FInCache[ square_pwm     ] := 0.5; // Set default PWM to 50% so it'll 'work'
      FInCache[ square_pwmlevel] := 1.0; // Set default PWM to 50% so it'll 'work'
      FKnobMap := FKnobMap + [ square_pwmlevel];
    end;

    procedure   TModSquare.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ baseosc_mute])
      then FOutCache[ baseosc_out] := 0
      else begin
        if FPosition <= ( 1.0 - FInCache[ square_pwm] * FInCache[ square_pwmlevel])
        then FOutCache[ baseosc_out] := -1.0
        else FOutCache[ baseosc_out] :=  1.0;
      end;
    end;

    function    TModSquare.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         square_pwm      : Result := 'pwm';
         square_pwmlevel : Result := 'pwmlevel';
        else               Result := inherited;
      end;
    end;


{ ========
  TModSquareTrig = class( TModTrigOsc)
  public
}

const

  squaretrig_pwm      =  9;
  squaretrig_pwmlevel = 10;


    procedure   TModSquareTrig.SetDefaults;
    begin
      inherited;
      FInCache[ squaretrig_pwm     ] := 0.5; // Set default PWM to 50% so it'll 'work'
      FInCache[ squaretrig_pwmlevel] := 1.0; // Set default PWM to 50% so it'll 'work'
      FKnobMap := FKnobMap + [ squaretrig_pwmlevel];
    end;

    procedure   TModSquareTrig.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ baseosc_mute])
      then FOutCache[ baseosc_out] := 0
      else begin
        if FPosition <= ( 1.0 - FInCache[ squaretrig_pwm] * FInCache[ squaretrig_pwmlevel])
        then FOutCache[ baseosc_out] := -1.0
        else FOutCache[ baseosc_out] :=  1.0;
      end;
    end;

    function    TModSquareTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         squaretrig_pwm      : Result := 'pwm';
         squaretrig_pwmlevel : Result := 'pwmlevel';
        else                   Result := inherited;
      end;
    end;


{ ========
  TModBaseLfo = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldRes   : Boolean;
    FRange    : Integer;
    FShape    : Integer;
  public
}

const

  baselfo_frequency = 0;
  baselfo_fm        = 1;
  baselfo_phase     = 2;
  baselfo_res       = 3;
  baselfo_mute      = 4;
  baselfo_fmlevel   = 5;
  baselfo_pmlevel   = 6;
  baselfo_shape     = 7;
  baselfo_range     = 8;

  baselfo_out       = 0;
  baselfo_sync      = 1;


    constructor TModBaseLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModBaseLfo.SetDefaults; // override;
    begin
      FKnobMap := [ baselfo_frequency, baselfo_fmlevel, baselfo_pmlevel];
    end;

    procedure   TModBaseLfo.DoSlowTick; // override;
    var
      Delta : TSignal;
    begin
      if ( not FOldRes) and SignalToLogic( FInCache[ baselfo_res]) or ResetFlag
      then FPhase := 0.0;
      ResetFlag := False;
      FOldRes   := SignalToLogic( FInCache[ baselfo_res]);

      FShape := SignalToInt( FInCache[ baselfo_shape]);
      FRange := SignalToInt( FInCache[ baselfo_range]);

      case FRange of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
        1  : Delta := LookupLfoMediumPhaseDelta( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
        else Delta := LookupLfoLinPhaseDelta   ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
      end;

      FPhase    := FModulo( FPhase + Delta, 1.0);
      FPosition := FModulo( FPhase + Clip( FInCache[ baselfo_phase] * FInCache[ baselfo_pmlevel], -1, 1), 1);
      FOutCache[ baselfo_sync] := LogicToSignal( FPosition < 0.01);
    end;

    procedure   TModBaseLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f' , [ aPrefix, Name, FInCache [ baselfo_res]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutCache[ baselfo_out]], AppLocale));
    end;

    function    TModBaseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baselfo_frequency : Result := 'frequency';
        baselfo_fm        : Result := 'fm';
        baselfo_phase     : Result := 'phase';
        baselfo_res       : Result := 'res';
        baselfo_mute      : Result := 'mute';
        baselfo_fmlevel   : Result := 'fmlevel';
        baselfo_pmlevel   : Result := 'pmlevel';
        baselfo_shape     : Result := 'shape';
        baselfo_range     : Result := 'range';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModBaseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baselfo_out  : Result := 'out';
        baselfo_sync : Result := 'sync';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseTrigLfo = class( TModLfo)
  private
    FState   : TTrigOscState;
    FCounter : TSignal;
  public
}

const

  basetriglfo_trig  = 2;
  basetriglfo_count = 3;

  basetriglfo_ready = 1;


    procedure   TModBaseTrigLfo.DoSlowTick; // override;
    var
      FOldCounter : Integer;
      Delta       : TSignal;
    begin
      if ResetFlag
      then begin
        FState    := toIdle;
        ResetFlag := False;
      end;

      case FState of

        toIdle :
          begin
            FPosition := 0.0;
            FCounter  := 0;
            if SignalToLogic( FInCache[ basetriglfo_trig])
            then FState := toRunning;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FShape := SignalToInt( FInCache[ baselfo_shape]);
            FRange := SignalToInt( FInCache[ baselfo_range]);

            case FRange of
              0  : Delta := LookupLfoFastPhaseDelta  ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
              1  : Delta := LookupLfoMediumPhaseDelta( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
              2  : Delta := LookupLfoSlowPhaseDelta  ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
              3  : Delta := LookupLfoBPMPhaseDelta   ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
              else Delta := LookupLfoLinPhaseDelta   ( FInCache[ baselfo_frequency] + FInCache[ baselfo_fm] * FInCache[ baselfo_fmlevel]);
            end;
            FPosition := FModulo( FPosition + Delta, 1.0);

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPosition, 0.5, Delta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPosition, 1.0, Delta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInCache[ basetriglfo_count]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPosition, 0.5, Delta) and
                not AlmostEqual( FPosition, 1.0, Delta)
              then FState := toRunning;
            end;

          end;

        toDone :
          begin
            FPosition := 0.0;
            if not SignalToLogic( FInCache[ basetriglfo_trig])
            then FState := toIdle;
          end;
      end;
      FOutCache[ basetriglfo_ready] := LogicToSignal( FState = toIdle);
    end;

    function    TModBaseTrigLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        basetriglfo_trig  : Result := 'trig';
        basetriglfo_count : Result := 'count';
        else                Result := inherited;
      end;
    end;

    function    TModBaseTrigLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        basetriglfo_ready : Result := 'ready';
        else                Result := inherited;
      end;
    end;

    procedure   TModBaseTrigLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInCache [ basetriglfo_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'  , [ aPrefix, Name, FOutCache[ baselfo_out      ]], AppLocale));
      aCallback( Self, Format( '%s/%s/ready/%f', [ aPrefix, Name, FOutCache[ basetriglfo_ready]], AppLocale));
    end;


{ ========
  TModLfo = class( TModBaseLfo)
  public
}

    procedure   TModLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ baselfo_mute])
      then FOutCache[ baselfo_out] := 0
      else begin
        case FShape of
          0  : // sine
            FOutCache[ baselfo_out] := LookupSine( 2 * FPosition);
          1  : // tri
            if FPosition <= 0.5
            then FOutCache[ baselfo_out] := AntiRunAway( 4.0 *         FPosition  - 1)
            else FOutCache[ baselfo_out] := AntiRunAway( 4.0 * ( 1.0 - FPosition) - 1);
          2  : // saw
            FOutCache[ baselfo_out] := AntiRunAway( 2 * FPosition - 1);
          3  : // square
            if FPosition < 0.5
            then FOutCache[ baselfo_out] := -1.0
            else FOutCache[ baselfo_out] :=  1.0;
        end;
      end;
    end;


{ ========
  TModLfoTrig = class( TModbaseTrigLfo)
  public
}

    procedure   TModLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ baselfo_mute])
      then FOutCache[ baselfo_out] := 0
      else begin
        case FShape of
          0  : // sine
            FOutCache[ baselfo_out] := LookupSine( 2 * FPosition);
          1  : // tri
            if FPosition <= 0.5
            then FOutCache[ baselfo_out] := AntiRunAway( 4.0 *         FPosition  - 1)
            else FOutCache[ baselfo_out] := AntiRunAway( 4.0 * ( 1.0 - FPosition) - 1);
          2  : // saw
            FOutCache[ baselfo_out] := AntiRunAway( 2 * FPosition - 1);
          3  : // square
            if FPosition < 0.5
            then FOutCache[ baselfo_out] := -1.0
            else FOutCache[ baselfo_out] :=  1.0;
        end;
      end;
    end;


{ ========
  TModSquareLfo = class( TModLfo)
  public
}

const

  squarelfo_pwm      =  9;
  squarelfo_pwmlevel = 10;


    procedure   TModSquareLfo.SetDefaults;
    begin
      inherited;
      FInCache[ squarelfo_pwm     ] := 0.5; // Set default PWM to 50% so it'll 'work'
      FInCache[ squarelfo_pwmlevel] := 1.0; // Set default PWM to 50% so it'll 'work'
      FKnobMap := [ baselfo_frequency, baselfo_fmlevel, baselfo_pmlevel, squarelfo_pwmlevel];
    end;

    procedure   TModSquareLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ baselfo_mute])
      then FOutCache[ baselfo_out] := 0
      else begin
        if FPosition <= ( 1.0 - FInCache[ squarelfo_pwm] * FInCache[ squarelfo_pwmlevel])
        then FOutCache[ baselfo_out] := -1.0
        else FOutCache[ baselfo_out] :=  1.0;
      end;
    end;

    function    TModSquareLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         squarelfo_pwm      : Result := 'pwm';
         squarelfo_pwmlevel : Result := 'pwmlevel';
        else                  Result := inherited;
      end;
    end;


{ ========
  TModTrigSquareLfo = class( TModBaseTrigLfo)
  public
}

const

  squarelfotrig_pwm      =  9;
  squarelfotrig_pwmlevel = 10;


    procedure   TModSquareLfoTrig.SetDefaults;
    begin
      inherited;
      FInCache[ squarelfotrig_pwm     ] := 0.5; // Set default PWM to 50% so it'll 'work'
      FInCache[ squarelfotrig_pwmlevel] := 1.0; // Set default PWM to 50% so it'll 'work'
      FKnobMap := [ baselfo_frequency, baselfo_fmlevel, baselfo_pmlevel, squarelfo_pwmlevel];
    end;

    procedure   TModSquareLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ baselfo_mute])
      then FOutCache[ baselfo_out] := 0
      else begin
        if FPosition <= ( 1.0 - FInCache[ squarelfotrig_pwm] * FInCache[ squarelfotrig_pwmlevel])
        then FOutCache[ baselfo_out] := -1.0
        else FOutCache[ baselfo_out] :=  1.0;
      end;
    end;

    function    TModSquareLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         squarelfo_pwm      : Result := 'pwm';
         squarelfo_pwmlevel : Result := 'pwmlevel';
        else                  Result := inherited;
      end;
    end;


{ ========
  TModRandomWalkLfo = class( TModLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
}

const

  randomwalklfo_frequency = 0;
  randomwalklfo_fm        = 1;
  randomwalklfo_distance  = 2;
  randomwalklfo_mute      = 3;
  randomwalklfo_color     = 4;

  randomwalklfo_outx      = 0;
  randomwalklfo_outy      = 1;
  randomwalklfo_outz      = 2;


    procedure   TModRandomWalkLfo.SetDefaults; // override;
    begin
      inherited;
      FKnobMap := [
        randomwalklfo_frequency,
        randomwalklfo_distance,
        randomwalklfo_mute,
        randomwalklfo_color
      ];
    end;

    procedure   TModRandomWalkLfo.DoSlowTick; // override;
    var
      Distance : TSignal;
      Delta    : TSignal;
      Color    : TSignal;
    begin

      Delta     := LookupLfoMediumPhaseDelta( FInCache[ randomwalklfo_frequency] + Clip( FInCache[ randomwalklfo_fm], -1, 1));
      FPosition := FModulo( FPosition + Delta, 1.0);

      case FState of
        trwIdle :
          begin
            if FPosition >= 0.5
            then begin
              FState   := trwFired;
              Distance := FInCache[ randomwalklfo_distance];
              case Random( 6) of
                0 : FX := ( FX + Distance) * SignalToMute( FInCache[ randomwalklfo_mute]);
                1 : FX := ( FX - Distance) * SignalToMute( FInCache[ randomwalklfo_mute]);
                2 : FY := ( FY + Distance) * SignalToMute( FInCache[ randomwalklfo_mute]);
                3 : FY := ( FY - Distance) * SignalToMute( FInCache[ randomwalklfo_mute]);
                4 : FZ := ( FZ + Distance) * SignalToMute( FInCache[ randomwalklfo_mute]);
                5 : FZ := ( FZ - Distance) * SignalToMute( FInCache[ randomwalklfo_mute]);
              end;
              if FX < -1 then FX := - 2 - FX else if FX > 1 then FX := 2 - FX;
              if FY < -1 then FY := - 2 - FY else if FY > 1 then FY := 2 - FY;
              if FZ < -1 then FZ := - 2 - FZ else if FZ > 1 then FZ := 2 - FZ;
              Color := FInCache[ randomwalklfo_color];
              FOutCache[ randomwalklfo_outx] := ( 1 - Color) * FOutCache[ randomwalklfo_outx] + Color * FX;
              FOutCache[ randomwalklfo_outy] := ( 1 - Color) * FOutCache[ randomwalklfo_outy] + Color * FY;
              FOutCache[ randomwalklfo_outz] := ( 1 - Color) * FOutCache[ randomwalklfo_outz] + Color * FZ;
            end;
          end;

        trwFired :
          begin
            if FPosition < 0.5
            then FState := trwIdle;
          end;
      end;
    end;

    procedure   TModRandomWalkLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/outx/%f', [ aPrefix, Name, FOutCache[ randomwalklfo_outx]], AppLocale));
      aCallback( Self, Format( '%s/%s/outy/%f', [ aPrefix, Name, FOutCache[ randomwalklfo_outy]], AppLocale));
      aCallback( Self, Format( '%s/%s/outz/%f', [ aPrefix, Name, FOutCache[ randomwalklfo_outz]], AppLocale));
    end;

    function    TModRandomWalkLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        randomwalklfo_frequency : Result := 'frequency';
        randomwalklfo_fm        : Result := 'fm';
        randomwalklfo_distance  : Result := 'distance';
        randomwalklfo_mute      : Result := 'mute';
        randomwalklfo_color     : Result := 'color';
        else Result := NO_INPUT;
      end;
    end;

    function    TModRandomWalkLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        randomwalklfo_outx : Result := 'outx';
        randomwalklfo_outy : Result := 'outy';
        randomwalklfo_outz : Result := 'outz';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoise = class( TMod)
  public
}

const

  noise_color = 0;
  noise_mute  = 1;

  noise_out   = 0;


    procedure   TModNoise.DoTick; // override;
    var
      Color : TSignal;
    begin
      Color := FInCache[ noise_color];
      if Color < -0.99
      then Color := -0.99;
      FOutCache[ noise_out] := AntiRunAway( FInCache[ noise_mute] * (( 1 - Color) * FOutCache[ noise_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color)));
    end;

    procedure   TModNoise.SetDefaults; // override;
    begin
      FKnobMap := [ noise_color, noise_mute];
    end;

    function    TModNoise.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noise_color : Result := 'color';
        noise_mute  : Result := 'mute';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModNoise.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noise_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrigNoise = class( TMod)
  private
    FLastInput: TSignal;
  public
}

const

  noisetrig_color = 0;
  noisetrig_trig  = 1;
  noisetrig_mute  = 2;

  noisetrig_out   = 0;


    procedure   TModNoiseTrig.DoTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInCache[ noisetrig_trig]);
      if not FLastTrig and Trig
      then begin
        Color := FInCache[ noisetrig_color];
        if Color < -0.99
        then Color := -0.99;
        FOutCache[ noisetrig_out] := AntiRunAway( FInCache[ noisetrig_mute] * (( 1 - Color) * FOutCache[ noisetrig_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color)));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseTrig.SetDefaults; // override;
    begin
      FKnobMap := [ noisetrig_color, noisetrig_mute];
    end;

    procedure   TModNoiseTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ noisetrig_trig]], AppLocale));
    end;

    function    TModNoiseTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noisetrig_color : Result := 'color';
        noisetrig_trig  : Result := 'trig';
        noisetrig_mute  : Result := 'mute';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModNoiseTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noisetrig_out : Result := 'out';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfo = class( TMod)
  public
}

const

  noiselfo_color = 0;
  noiselfo_mute  = 1;

  noiselfo_out   = 0;


    constructor TModNoiseLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModNoiseLfo.SetDefaults; // override;
    begin
      FKnobMap := [ noiselfo_color, noiselfo_mute];
    end;

    procedure   TModNoiseLfo.DoSlowTick; // override;
    var
      Color : TSignal;
    begin
      Color := FInCache[ noiselfo_color];
      if Color < -0.99
      then Color := -0.99;
      FOutCache[ noiselfo_out] := AntiRunAway( FInCache[ noiselfo_mute] * (( 1 - Color) * FOutCache[ noiselfo_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color)));
    end;

    function    TModNoiseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfo_color : Result := 'color';
        noiselfo_mute  : Result := 'mute';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfo_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfoTrig = class( TMod)
  private
    FLastInput: TSignal;
  public
}

const

  noiselfotrig_color = 0;
  noiselfotrig_trig  = 1;
  noiselfotrig_mute  = 2;

  noiselfotrig_out   = 0;


    constructor TModNoiseLfoTrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModNoiseLfoTrig.SetDefaults; // override;
    begin
      FKnobMap := [ noiselfotrig_color, noiselfotrig_mute];
    end;

    procedure   TModNoiseLfoTrig.DoSlowTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInCache[ noiselfotrig_trig]);
      if not FLastTrig and Trig
      then begin
        Color := FInCache[ noiselfotrig_color];
        if Color < -0.99
        then Color := -0.99;
        FOutCache[ noiselfotrig_out] := AntiRunAway( FInCache[ noiselfotrig_mute] * (( 1 - Color) * FOutCache[ noiselfotrig_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color)));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseLfoTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ noiselfotrig_trig]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutCache[ noiselfotrig_out ]], AppLocale));
    end;

    function    TModNoiseLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfotrig_color : Result := 'color';
        noiselfotrig_trig  : Result := 'trig';
        noiselfotrig_mute  : Result := 'mute';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfoTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfotrig_out : Result := 'out';
        else Result      := NO_OUTPUT;
      end;
    end;


{ ========
  TModAvg = class( TMod)
  private
    FAlpha : TSignal;
  public
}

const

  avg_in    = 0;
  avg_alpha = 1;

  avg_lp    = 0;
  avg_hp    = 1;


    constructor TModAverage.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
    end;

    procedure   TModAverage.SetDefaults; // override;
    begin
      FKnobMap := [ avg_alpha];
    end;

    procedure   TModAverage.DoSlowTick; // override;
    begin
      FAlpha := AntiRunAway( ExpAttack( FInCache[ avg_alpha], 1, 0, 1));
    end;

    procedure   TModAverage.DoTick; // override;
    var
      Inp  : TSignal;
      Outp : TSignal;
    begin
      Inp   := AntiRunAway( FInCache[ avg_in]);
      Outp  := AntiRunAway(( 1 -  FAlpha) * FOutCache[ avg_lp] + FAlpha * Inp);
      FOutCache[ avg_lp] := Outp;
      FOutCache[ avg_hp] := Inp - Outp;
    end;

    function    TModAverage.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        avg_in    : Result := 'in';
        avg_alpha : Result := 'alpha';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModAverage.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        avg_lp : Result := 'lp';
        avg_hp : Result := 'hp';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSVF = class( TMod)
  private
    FFreq  : TSignal;
    FDamp  : TSignal;
    FDrive : TSignal;
  public
}

    (*
      from    : http://www.musicdsp.org/archive.php?classid=3#92
      also see: http://www.music.mcgill.ca/~ich/classes/FiltersChap2.pdf

      input  = input buffer;
      output = output buffer;
      fs     = sampling frequency;
      fc     = cutoff frequency normally something like:
               440.0*pow(2.0, (midi_note - 69.0)/12.0);                 :: We use signals rather, Frequency = Signal * SampleRate / 2
      res    = resonance 0 to 1;
      drive  = internal distortion 0 to 0.1
      freq   = 2.0*sin(PI*MIN(0.25, fc/(fs*2)));  // the fs*2 is because it's double sampled
      damp   = MIN(2.0*(1.0 - pow(res, 0.25)), MIN(2.0, 2.0/freq - freq*0.5));
      notch  = notch output
      low    = low pass output
      high   = high pass output
      band   = band pass output
      peak   = peaking output = low - high
      --
      double sampled svf loop:
      for (i=0; i<numSamples; i++)
      {
        in    = input[i];
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out   = 0.5*(notch or low or high or band or peak);
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out  += 0.5*(same out as above);
        output[i] = out;
      }
    *)

const

  svf_in         = 0;
  svf_frequency  = 1;
  svf_q          = 2;
  svf_distortion = 3;
  svf_fm         = 4;
  svf_fmlevel    = 5;
  svf_inlevel    = 6;

  svf_lp         = 0;
  svf_bp         = 1;
  svf_hp         = 2;
  svf_br         = 3;


    constructor TModSVF.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
    end;

    procedure   TModSVF.SetDefaults; // override;
    begin
      FKnobMap := [
        svf_frequency,
        svf_q,
        svf_distortion,
        svf_fmlevel,
        svf_inlevel
      ];
    end;

    procedure   TModSVF.DoSlowTick; // override;
    // var
    //   Fc : TSignal;
    begin
      //  Fc    := LookupFrequency( FInCache[ svf_frequency] + FInCache[ svf_fm]);
      //  FFreq := 2 * Sin( PI * Min( 0.25, ( Fc / SYSTEM_RATE_2)));

      // FFreq := UnitsToSvfFrequency( FInCache[ svf_frequency] + FInCache[ svf_fm]);
      FFreq := LookupSvfFrequency( FInCache[ svf_frequency] + FInCache[ svf_fm] * FInCache[ svf_fmlevel]);

      // damp   = MIN( 2.0 * ( 1.0 - pow( res, 0.25)), MIN( 2.0, 2.0 / freq - freq * 0.5));

      if FFreq <= 0.0001
      then FDamp := Min( 2.0 * ( 1.0 - Power( FInCache[ svf_q], 0.25)), 2.0)
      else FDamp := Min( 2.0 * ( 1.0 - Power( FInCache[ svf_q], 0.25)), Min( 2.0, 2.0 / FFreq - FFreq * 0.5));

      FDrive := Clip( FInCache[ svf_distortion] / 10, 0.0, 0.1);
    end;

    procedure   TModSVF.DoTick; // override;
    const
      FOVER = 2; // over sampling rate.
    var
      HP    : TSignal;
      BP    : TSignal;
      LP    : TSignal;
      BR    : TSignal;
      OutHP : TSignal;
      OutBP : TSignal;
      OutLP : TSignal;
      OutBR : TSignal;
      i     : Integer;
    begin
      LP    := FOutCache[ svf_lp];
      OutHP := 0.0;
      OutBP := 0.0;
      OutLP := 0.0;
      OutBR := 0.0;
      BP    := FOutCache[ svf_bp];

      for i := 1 to FOVER
      do begin
        BR    := FInCache[ svf_in] * FInCache[ svf_inlevel] - FDamp * BP;
        LP    := LP + FFreq * BP;
        HP    := BR - LP;
        BP    := FFreq * HP + BP - FDrive * BP * BP * BP; // with 0 <= Drive <= 0.1
        OutHP := OutHP + HP;
        OutBP := OutBP + BP;
        OutLP := OutLP + LP;
        OutBR := OutBR + BR;
      end;

      FOutCache[ svf_lp] := AntiRunAway( OutLP / FOVER);
      FOutCache[ svf_bp] := AntiRunAway( OutBP / FOVER);
      FOutCache[ svf_hp] := AntiRunAway( OutHP / FOVER);
      FOutCache[ svf_br] := AntiRunAway( OutBR / FOVER);
    end;

    function    TModSVF.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        svf_in         : Result := 'in';
        svf_frequency  : Result := 'frequency';
        svf_q          : Result := 'q';
        svf_distortion : Result := 'distortion';
        svf_fm         : Result := 'fm';
        svf_fmlevel    : Result := 'fmlevel';
        svf_inlevel    : Result := 'inlevel';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModSVF.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        svf_lp : Result := 'lp';
        svf_bp : Result := 'bp';
        svf_hp : Result := 'hp';
        svf_br : Result := 'br';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
  public
}

const

  envar_trig    = 0;
  envar_attack  = 1;
  envar_release = 2;
  envar_gain    = 3;
  envar_sigin   = 4;
  envar_mute    = 5;

  envar_out     = 0;
  envar_sigout  = 1;


    procedure   TModEnvAR.SetDefaults; // override;
    begin
      inherited;
      FInCache[ envar_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FKnobMap := [ envar_attack, envar_release];
    end;

    procedure   TModEnvAR.DoTick; // override;
    var
      R : TSignal;
    begin
      if ResetFlag
      then begin
        FState    := arIdle;
        ResetFlag := False;
      end;

      case FState of

        arIdle :

          if SignalToLogic( FInCache[ envar_trig])
          then begin
            if FInCache[ envar_attack] <= 0
            then FCoeffAttack := 1
            else FCoeffAttack := 1 / Max( 1, TimeToSampleCount( FInCache[ envar_attack])); // Linear attack
            if FInCache[ envar_release] <= 0
            then FCoeffRelease := -1
            else FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / Max( 1, TimeToSampleCount( FInCache[ envar_release])); // Exp release
            // if FCoeffRelease > 0.999
            // then FCoeffRelease := 0.999;
            FState := arA;
          end;

        arA :
          begin
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
            case SignalToInt( FInCache[ envar_mute]) of
              1  : FOutCache[ envar_out] := 0;
              2  : FOutCache[ envar_out] := 1;
              else FOutCache[ envar_out] := AntiRunAway( FValue * FInCache[ envar_gain]);
            end;
          end;

        arR :
          begin
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
            case SignalToInt( FInCache[ envar_mute]) of
              1  : FOutCache[ envar_out] := 0;
              2  : FOutCache[ envar_out] := 1;
              else FOutCache[ envar_out] := AntiRunAway( FValue * FInCache[ envar_gain]);
            end;
          end;

        arWait :
          if not SignalToLogic( FInCache[ envar_trig])
          then FState := arIdle;

      end;
      case SignalToInt( FInCache[ envar_mute]) of
        1  : FOutCache[ envar_sigout] := 0;
        2  : FOutCache[ envar_sigout] := AntiRunAway( FInCache[ envar_sigin]);
        else FOutCache[ envar_sigout] := AntiRunAway( FInCache[ envar_sigin] * FOutCache[ envar_out]);
      end;
    end;

    procedure   TModEnvAR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ envar_trig]], AppLocale));
    end;

    function    TModEnvAR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_trig    : Result := 'trig';
        envar_attack  : Result := 'attack';
        envar_release : Result := 'release';
        envar_gain    : Result := 'gain';
        envar_sigin   : Result := 'sigin';
        envar_mute    : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModEnvAR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_out    : Result := 'out';
        envar_sigout : Result := 'sigout';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
  public
}

    procedure   TModEnvARRetrig.SetDefaults; // override;
    begin
      inherited;
      FInCache[ envar_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FKnobMap := [ envar_attack, envar_release];
    end;

    procedure   TModEnvARRetrig.DoTick; // override;
    var
      R : TSignal;
    begin
      if ResetFlag
      then begin
        FState    := arIdle;
        ResetFlag := False;
      end;

      if FArmed and SignalToLogic( FInCache[ envar_trig])
      then FState := arIdle;

      case FState of

        arIdle :

          if SignalToLogic( FInCache[ envar_trig])
          then begin
            FCoeffAttack  := 1 / Max( 1, TimeToSampleCount( FInCache[ envar_attack])); // Linear attack
            FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / Max( 1, TimeToSampleCount( FInCache[ envar_release])); // Exp release
            FState        := arA;
            FArmed        := False;
          end;

        arA :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInCache[ envar_trig]);
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
            case SignalToInt( FInCache[ envar_mute]) of
              1 :  FOutCache[ envar_out] := 0;
              2 :  FOutCache[ envar_out] := 1;
              else FOutCache[ envar_out] := AntiRunAway( FValue * FInCache[ envar_gain]);
            end;
          end;

        arR :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInCache[ envar_trig]);
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
            case SignalToInt( FInCache[ envar_mute]) of
              1 :  FOutCache[ envar_out] := 0;
              2 :  FOutCache[ envar_out] := 1;
              else FOutCache[ envar_out] := AntiRunAway( FValue * FInCache[ envar_gain]);
            end;
          end;

        arWait :
          if not SignalToLogic( FInCache[ envar_trig])
          then FState := arIdle;

      end;
      case SignalToInt( FInCache[ envar_mute]) of
        1 :  FOutCache[ envar_sigout] := 0;
        2 :  FOutCache[ envar_sigout] := AntiRunAway( FInCache[ envar_sigin]);
        else FOutCache[ envar_sigout] := AntiRunAway( FInCache[ envar_sigin] * FOutCache[ envar_out]);
      end;
    end;

    procedure   TModEnvARRetrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ envar_trig]], AppLocale));
    end;

    function    TModEnvARRetrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_trig    : Result := 'trig';
        envar_attack  : Result := 'attack';
        envar_release : Result := 'release';
        envar_gain    : Result := 'gain';
        envar_sigin   : Result := 'sigin';
        envar_mute    : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModEnvARRetrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_out    : Result := 'out';
        envar_sigout : Result := 'sigout';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAHD = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TAHDState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
  public
}

const

  envahd_trig      =  0;
  envahd_mode      =  1;
  envahd_ashape    =  2;
  envahd_dshape    =  3;
  envahd_attack    =  4;
  envahd_hold      =  5;
  envahd_decay     =  6;
  envahd_gain      =  7;
  envahd_sigin     =  8;
  envahd_mute      =  9;
  envahd_arange    = 10;
  envahd_hrange    = 11;
  envahd_drange    = 12;
  envahd_amod      = 13;
  envahd_hmod      = 14;
  envahd_dmod      = 15;
  envahd_amodneg   = 16;
  envahd_hmodneg   = 17;
  envahd_dmodneg   = 18;
  envahd_amodlevel = 19;
  envahd_hmodlevel = 20;
  envahd_dmodlevel = 21;

  envahd_out       =  0;
  envahd_sigout    =  1;

    constructor TModEnvAHD.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
    end;

    procedure   TModEnvAHD.SetDefaults; // override;
    begin
      FInCache[ envahd_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FKnobMap := [
        envahd_attack,
        envahd_hold,
        envahd_decay,
        envahd_amodlevel,
        envahd_hmodlevel,
        envahd_dmodlevel
      ];
    end;

    procedure   TModEnvAHD.DoTick; // override;
    var
      R      : TSignal;
      Trig   : Boolean;
      Mode   : Integer;   // Gate / Trig
      AShape : TEnvShape;
      DShape : TEnvShape;
      Gain   : TSignal;
      AMod   : TSignal;
      HMod   : TSignal;
      DMod   : TSignal;
    begin
      if ResetFlag
      then begin
        FState    := ahdIdle;
        ResetFlag := False;
      end;

      Trig    := SignalToLogic   ( FInCache[ envahd_trig  ]);
      Mode    := SignalToInt     ( FInCache[ envahd_mode  ]);
      AShape  := SignalToEnvShape( FInCache[ envahd_ashape]);
      DShape  := SignalToEnvShape( FInCache[ envahd_dshape]);
      Gain    :=                   FInCache[ envahd_gain  ];
      R       := FValue;

      if FArmed and Trig and ( Mode = 1)
      then FState := ahdIdle;

      case FState of

        ahdIdle :

          if Trig
          then begin
            FState      := ahdA;
            FArmed      := False;
            FLastTarget := FValue;
            AMod        := ( FInCache[ envahd_amod] - FInCache[ envahd_amodneg]) * FInCache[ envahd_amodlevel];
            FTime       := 0;
            FDuration   := Round( TimeToSampleCount( FInCache[ envahd_attack] * ( 1 + AMod)));
          end;

        ahdA :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case AShape of
              esLog  : R := LogAttack( FTime, FDuration, FLastTarget, 1);
              esLin  : R := LinAttack( FTime, FDuration, FLastTarget, 1);
              else     R := ExpAttack( FTime, FDuration, FLastTarget, 1);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FTime       := 0;
              HMod        := ( FInCache[ envahd_hmod] - FInCache[ envahd_hmodneg]) * FInCache[ envahd_hmodlevel];
              FDuration   := Round( TimeToSampleCount( FInCache[ envahd_hold] * ( 1 + HMod)));
              FState      := ahdH;
            end;
          end;

        ahdH :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FTime       := 0;
              DMod        := ( FInCache[ envahd_dmod] - FInCache[ envahd_dmodneg]) * FInCache[ envahd_dmodlevel];
              FDuration   := Round( TimeToSampleCount( FInCache[ envahd_decay] * ( 1 + DMod)));
              FState      := ahdD;
            end;
          end;

        ahdD :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case DShape of
              esLog : R := LogDecay( FTime, FDuration, FLastTarget, 0);
              esLin : R := LinDecay( FTime, FDuration, FLastTarget, 0);
              else    R := ExpDecay( FTime, FDuration, FLastTarget, 0);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FState      := ahdWait;
            end;
          end;

        ahdWait :
          if not Trig or FArmed
          then FState := ahdIdle;

      end;
      FValue        := R;
      case SignalToInt( FInCache[ envahd_mute]) of
        1 : begin
          FOutCache[ envahd_out   ] := 0;
          FOutCache[ envahd_sigout] := 0;
        end;
        2 : begin
          FOutCache[ envahd_out   ] := 1;
          FOutCache[ envahd_sigout] := AntiRunAway( FInCache[ envahd_sigin]);
        end;
        else begin
          FOutCache[ envahd_out   ] := AntiRunAway( FValue * Gain);
          FOutCache[ envahd_sigout] := AntiRunAway( FInCache[ envahd_sigin] * FValue * Gain); // sigout := sigin * out
        end;
      end;
    end;

    procedure   TModEnvAHD.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ envahd_trig]], AppLocale));
    end;

    procedure   TModEnvAHD.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime  : TSignal;
      HTime  : TSignal;
      DTime  : TSignal;
      TTime  : TSignal;
      i      : Integer;
      p      : Integer;
      AShape : TEnvShape;
      DShape : TEnvShape;
      R      : TSignal;
    begin
      ATime  := FInCache[ envahd_attack];
      HTime  := FInCache[ envahd_hold  ];
      DTime  := FInCache[ envahd_decay ];
      TTime  := ATime + HTime + DTime;

      if TTIme = 0
      then Exit;

      ATime  := ATime * 67 / TTime;
      HTime  := HTime * 67 / TTime;
      DTime  := DTime * 67 / TTime;
      AShape := SignalToEnvShape( FInCache[ envahd_ashape]);
      DShape := SignalToEnvShape( FInCache[ envahd_dshape]);

      i := 0;
      while i < ATime
      do begin
        case AShape of
          esLog  : R := LogAttack( i, ATime, 0, 1);
          esLin  : R := LinAttack( i, ATime, 0, 1);
          else     R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] := R;
        Inc( i);
      end;
      while i < ATime + HTime
      do begin
        FData[ i] := 1;
        Inc( i);
      end;
      p := 0;
      while i < 67
      do begin
        case DShape of
          esLog : R := LogDecay( p, DTime, 1, 0);
          esLin : R := LinDecay( p, DTime, 1, 0);
          else    R := ExpDecay( p, DTime, 1, 0);
        end;
        FData[ i] := R;
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvAHD.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envahd_trig      : Result := 'trig';
        envahd_mode      : Result := 'mode';
        envahd_ashape    : Result := 'ashape';
        envahd_dshape    : Result := 'dshape';
        envahd_attack    : Result := 'attack';
        envahd_hold      : Result := 'hold';
        envahd_decay     : Result := 'decay';
        envahd_gain      : Result := 'gain';
        envahd_sigin     : Result := 'sigin';
        envahd_mute      : Result := 'mute';
        envahd_arange    : Result := 'arange';
        envahd_hrange    : Result := 'hrange';
        envahd_drange    : Result := 'drange';
        envahd_amod      : Result := 'amod';
        envahd_hmod      : Result := 'hmod';
        envahd_dmod      : Result := 'dmod';
        envahd_amodneg   : Result := 'amodneg';
        envahd_hmodneg   : Result := 'hmodneg';
        envahd_dmodneg   : Result := 'dmodneg';
        envahd_amodlevel : Result := 'amodlevel';
        envahd_hmodlevel : Result := 'hmodlevel';
        envahd_dmodlevel : Result := 'dmodlevel';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModEnvAHD.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envahd_out    :  Result := 'out';
        envahd_sigout :  Result := 'sigout';
        else             Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModEnvADSR = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TADSRState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
  public
}

const

  envadsr_trig      =  0;
  envadsr_mode      =  1;
  envadsr_ashape    =  2;
  envadsr_dshape    =  3;
  envadsr_rshape    =  4;
  envadsr_attack    =  5;
  envadsr_decay     =  6;
  envadsr_sustain   =  7;
  envadsr_release   =  8;
  envadsr_gain      =  9;
  envadsr_sigin     = 10;
  envadsr_mute      = 11;
  envadsr_arange    = 12;
  envadsr_drange    = 13;
  envadsr_rrange    = 14;
  envadsr_amod      = 15;
  envadsr_dmod      = 16;
  envadsr_smod      = 17;
  envadsr_rmod      = 18;
  envadsr_amodneg   = 19;
  envadsr_dmodneg   = 20;
  envadsr_smodneg   = 21;
  envadsr_rmodneg   = 22;
  envadsr_amodlevel = 23;
  envadsr_dmodlevel = 24;
  envadsr_smodlevel = 25;
  envadsr_rmodlevel = 26;

  envadsr_out       =  0;
  envadsr_sigout    =  1;

    constructor TModEnvADSR.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
    end;

    procedure   TModEnvADSR.SetDefaults; // override;
    begin
      inherited;
      FInCache[ envadsr_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FKnobMap := [
        envadsr_attack,
        envadsr_decay,
        envadsr_sustain,
        envadsr_release,
        envadsr_amodlevel,
        envadsr_dmodlevel,
        envadsr_smodlevel,
        envadsr_rmodlevel
      ];
    end;

    procedure   TModEnvADSR.DoTick; // override;
    var
      R      : TSignal;
      Trig   : Boolean;
      Mode   : Integer; // Gate / Trig
      AMod   : TSignal;
      DMod   : TSignal;
      SMod   : TSignal;
      RMod   : TSignal;
      Gain   : TSignal;
    begin
      if ResetFlag
      then begin
        FState    := adsrIdle;
        ResetFlag := False;
      end;

      Trig    := SignalToLogic( FInCache[ envadsr_trig]);
      Mode    := SignalToInt  ( FInCache[ envadsr_mode]);
      Gain    :=                FInCache[ envadsr_gain];
      R       := FValue;

      if FArmed and Trig and ( Mode = 1)
      then FState := adsrIdle;

      case FState of

        adsrIdle :

          if Trig
          then begin
            FState      := adsrA;
            FArmed      := False;
            FLastTarget := FValue;
            FTime       := 0;
            AMod        := ( FInCache[ envadsr_amod] - FInCache[ envadsr_amodneg]) * FInCache[ envadsr_amodlevel];
            FDuration   := Round( TimeToSampleCount( FInCache[ envadsr_attack] * ( 1 + AMod)));
          end;

        adsrA :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case SignalToEnvShape( FInCache[ envadsr_ashape ]) of
              esLog : R := LogAttack( FTime, FDuration, FLastTarget, 1);
              esLin : R := LinAttack( FTime, FDuration, FLastTarget, 1);
              else    R := ExpAttack( FTime, FDuration, FLastTarget, 1);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FTime       := 0;
              DMod        := ( FInCache[ envadsr_amod] - FInCache[ envadsr_amodneg]) * FInCache[ envadsr_amodlevel];
              FDuration   := Round( TimeToSampleCount( FInCache[ envadsr_decay] * ( 1 + DMod)));
              FState      := adsrD;
            end;
          end;

        adsrD :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            SMod := ( FInCache[ envadsr_amod] - FInCache[ envadsr_amodneg]) * FInCache[ envadsr_amodlevel];
            case SignalToEnvShape( FInCache[ envadsr_dshape ]) of
              esLog : R := LogDecay( FTime, FDuration, FLastTarget, FInCache[ envadsr_sustain] + SMod);
              esLin : R := LinDecay( FTime, FDuration, FLastTarget, FInCache[ envadsr_sustain] + SMod);
              else    R := ExpDecay( FTime, FDuration, FLastTarget, FInCache[ envadsr_sustain] + SMod);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FState      := adsrS;
            end;
          end;

        adsrS :
          begin
            if not FArmed
            then FArmed := not Trig;
            if FArmed
            then begin
              FTime       := 0;
              RMod        := ( FInCache[ envadsr_amod] - FInCache[ envadsr_amodneg]) * FInCache[ envadsr_amodlevel];
              FDuration   := Round( TimeToSampleCount( FInCache[ envadsr_release] * ( 1 + RMod)));
              FState      := adsrR;
            end;
          end;

        adsrR :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case SignalToEnvShape( FInCache[ envadsr_rshape ]) of
              esLog : R := LogDecay( FTime, FDuration, FLastTarget, 0);
              esLin : R := LinDecay( FTime, FDuration, FLastTarget, 0);
              else    R := ExpDecay( FTime, FDuration, FLastTarget, 0);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FState      := adsrWait;
            end;
          end;

        adsrWait :
          if not Trig or FArmed
          then FState := adsrIdle;

      end;
      FValue := R;
      case SignalToInt( FInCache[ envadsr_mute]) of
        1 : begin
          FOutCache[ envadsr_out   ] := 0;
          FOutCache[ envadsr_sigout] := 0;
        end;
        2 : begin
          FOutCache[ envadsr_out   ] := 1;
          FOutCache[ envadsr_sigout] := AntiRunAway( FInCache[ envadsr_sigin]);
        end;
        else begin
          FOutCache[ envadsr_out   ] := AntiRunAway( FValue * Gain);
          FOutCache[ envadsr_sigout] := AntiRunAway( FInCache[ envadsr_sigin] * FValue * Gain); // sigout := sigin * out
        end;
      end;
    end;

    procedure   TModEnvADSR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ envadsr_trig]], AppLocale));
    end;

    procedure   TModEnvADSR.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime  : TSignal;
      DTime  : TSignal;
      STime  : TSignal;
      RTime  : TSignal;
      TTime  : TSignal;
      SLevel : TSignal;
      i      : Integer;
      p      : Integer;
      R      : TSignal;
    begin
      ATime  := FInCache[ envadsr_attack ];
      DTime  := FInCache[ envadsr_decay  ];
      SLevel := FInCache[ envadsr_sustain];
      RTime  := FInCache[ envadsr_release];
      STime  := ( ATime + DTime + RTime) / 3;
      TTime  := ATime + DTime + STime + RTime;

      if TTIme = 0
      then Exit;

      ATime  := ATime * 67 / TTime;
      DTime  := DTime * 67 / TTime;
      RTime  := RTime * 67 / TTime;
      STime  := STime * 67 / TTime;

      i := 0;
      while i < ATime
      do begin
        case SignalToInt( FInCache[ envadsr_ashape]) of
          0  : R := LogAttack( i, ATime, 0, 1);
          1  : R := LinAttack( i, ATime, 0, 1);
          else R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] := R;
        Inc( i);
      end;

      p := 0;
      while i < ATime + DTime
      do begin
        case SignalToEnvShape( FInCache[ envadsr_dshape]) of
          esLog : R := LogDecay( p, DTime, 1, SLevel);
          esLin : R := LinDecay( p, DTime, 1, SLevel);
          else    R := ExpDecay( p, DTime, 1, SLevel);
        end;
        FData[ i] := R;
        Inc( i);
        Inc( p);
      end;

      while i < ATime + DTime + STime
      do begin
        FData[ i] := SLevel;
        Inc( i);
      end;

      p := 0;
      while i < 67
      do begin
        case SignalToEnvShape( FInCache[ envadsr_rshape]) of
          esLog : R := LogDecay( p, RTime, SLevel, 0);
          esLin : R := LinDecay( p, RTime, SLevel, 0);
          else R := ExpDecay( p, RTime, SLevel, 0);
        end;
        FData[ i] := R;
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvADSR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envadsr_trig      : Result := 'trig';
        envadsr_mode      : Result := 'mode';
        envadsr_ashape    : Result := 'ashape';
        envadsr_dshape    : Result := 'dshape';
        envadsr_rshape    : Result := 'rshape';
        envadsr_attack    : Result := 'attack';
        envadsr_decay     : Result := 'decay';
        envadsr_sustain   : Result := 'sustain';
        envadsr_release   : Result := 'release';
        envadsr_gain      : Result := 'gain';
        envadsr_sigin     : Result := 'sigin';
        envadsr_mute      : Result := 'mute';
        envadsr_arange    : Result := 'arange';
        envadsr_drange    : Result := 'drange';
        envadsr_rrange    : Result := 'rrange';
        envadsr_amod      : Result := 'amod';
        envadsr_dmod      : Result := 'dmod';
        envadsr_smod      : Result := 'smod';
        envadsr_rmod      : Result := 'rmod';
        envadsr_amodneg   : Result := 'amodneg';
        envadsr_dmodneg   : Result := 'dmodneg';
        envadsr_smodneg   : Result := 'smodneg';
        envadsr_rmodneg   : Result := 'rmodneg';
        envadsr_amodlevel : Result := 'amodlevel';
        envadsr_dmodlevel : Result := 'dmodlevel';
        envadsr_smodlevel : Result := 'smodlevel';
        envadsr_rmodlevel : Result := 'rmodlevel';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModEnvADSR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envadsr_out    :  Result := 'out';
        envadsr_sigout :  Result := 'sigout';
        else              Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps   : Integer;
    FCounter : Integer;
    FInState : TSeqInState;
    FOldRes  : Boolean;
  public
}

const

  sequencer_trig   =  0;
  sequencer_reset  =  1;
  sequencer_dir    =  2;
  sequencer_chain  =  3;
  sequencer_step1  =  4;   // Keep steps consecutive
  sequencer_step2  =  5;
  sequencer_step3  =  6;
  sequencer_step4  =  7;
  sequencer_step5  =  8;
  sequencer_step6  =  9;
  sequencer_step7  = 10;
  sequencer_step8  = 11;
  sequencer_step9  = 12;
  sequencer_step10 = 13;
  sequencer_step11 = 14;

  sequencer_out    =  0;   // Keep trigs comsecutive
  sequencer_trig1  =  1;
  sequencer_trig2  =  2;
  sequencer_trig3  =  3;
  sequencer_trig4  =  4;
  sequencer_trig5  =  5;
  sequencer_trig6  =  6;
  sequencer_trig7  =  7;
  sequencer_trig8  =  8;
  sequencer_trig9  =  9;
  sequencer_trig10 = 10;
  sequencer_trig11 = 11;


    constructor TModSequencer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModSequencer.SetDefaults; // override;
    begin
      inherited;
      FOutValue                  := 0;
      FInCache[ sequencer_dir  ] := 1; // Make seq run forward by default.
      FInCache[ sequencer_chain] := 0; // Chain input to zero
    end;

    procedure   TModSequencer.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 11)
      else inherited;
    end;

    function    TModSequencer.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSequencer.DoSlowTick; // override;
    var
      i  : Integer;
    begin
      if ResetFlag
      then begin
        FInState  := seqIdle;
        FCounter  := 0;
        ResetFlag := False;
        FOutValue := 0;
      end;

      if not FOldRes and SignalToLogic( FInCache[ sequencer_reset])
      then begin
        FInState := seqIdle;
        FCounter := 0;
      end;
      FOldRes := SignalToLogic( FInCache[ sequencer_reset]);

      case FInState of
      seqIdle :
        begin
          if SignalToLogic( FInCache[ sequencer_trig])
          then begin
            FInState := seqHigh;
            if SignalToLogic( FInCache[ sequencer_dir])
            then FCounter := ( FCounter + 1) mod FSteps
            else begin
              if FCounter <= 0
              then FCounter := FSteps - 1
              else Dec( FCounter);
            end;
            FOutValue := FInCache[ FCounter + sequencer_step1] - MiddleNote / 128;
          end;
        end;
      seqHigh :
        begin
          if FInCache[ sequencer_trig] <= 0
          then FInState := seqIdle;
        end;
      end;

      FOutCache[ sequencer_out] := FOutValue + FInCache[ sequencer_chain];

      for i := sequencer_trig1 to sequencer_trig11
      do begin
        if i = FCounter + sequencer_trig1
        then FOutCache[ i] := LogicToSignal( True )
        else FOutCache[ i] := LogicToSignal( False);
      end;
    end;

    procedure   TModSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInCache [ sequencer_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInCache [ sequencer_reset]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInCache [ sequencer_dir  ]], AppLocale));
      for i := sequencer_trig1 to sequencer_trig11
      do begin
        if i = FCounter + sequencer_trig1
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - sequencer_trig1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - sequencer_trig1 + 1], AppLocale));
      end;
    end;

    function    TModSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sequencer_trig   : Result := 'trig';
        sequencer_reset  : Result := 'reset';
        sequencer_dir    : Result := 'dir';
        sequencer_chain  : Result := 'chain';
        sequencer_step1  : Result := 'step1';
        sequencer_step2  : Result := 'step2';
        sequencer_step3  : Result := 'step3';
        sequencer_step4  : Result := 'step4';
        sequencer_step5  : Result := 'step5';
        sequencer_step6  : Result := 'step6';
        sequencer_step7  : Result := 'step7';
        sequencer_step8  : Result := 'step8';
        sequencer_step9  : Result := 'step9';
        sequencer_step10 : Result := 'step10';
        sequencer_step11 : Result := 'step11';
        else               Result := NO_INPUT;;
      end;
    end;

    function    TModSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sequencer_out    : Result := 'out';
        sequencer_trig1  : Result := 'trig1';
        sequencer_trig2  : Result := 'trig2';
        sequencer_trig3  : Result := 'trig3';
        sequencer_trig4  : Result := 'trig4';
        sequencer_trig5  : Result := 'trig5';
        sequencer_trig6  : Result := 'trig6';
        sequencer_trig7  : Result := 'trig7';
        sequencer_trig8  : Result := 'trig8';
        sequencer_trig9  : Result := 'trig9';
        sequencer_trig10 : Result := 'trig10';
        sequencer_trig11 : Result := 'trig11';
        else               Result := NO_OUTPUT;
      end;
    end;

    function    TModSequencer.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModTextSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
}

const

  textsequencer_trig  = 0;
  textsequencer_reset = 1;
  textsequencer_dir   = 2;
  textsequencer_mode  = 3;
  textsequencer_chain = 4;
  textsequencer_step1 = 5;

  textsequencer_out   = 0;


    constructor TModTextSequencer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModTextSequencer.SetDefaults; // override;
    begin
      inherited;
      FOutValue                      := 0;
      FInCache[ textsequencer_dir  ] := 1; // Make seq run forward by default.
      FInCache[ textsequencer_chain] := 0; // Clear chain input
    end;

    procedure   TModTextSequencer.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        FInState  := seqIdle;
        FCounter  := 0;
        ResetFlag := False;
        FOutValue := 0;
        FSteps    := 0;
      end;

      if not FOldRes and SignalToLogic( FInCache[ textsequencer_reset])
      then begin
        FInState := seqIdle;
        FCounter := 0;
      end;

      case FInState of

        seqIdle :
          begin
            if SignalToLogic( FInCache[ textsequencer_trig])
            then begin
              FInState := seqHigh;
              if FInCache[ textsequencer_mode] = 0
              then begin
                if SignalToLogic( FInCache[ textsequencer_dir])
                then FCounter := ( FCounter + 1) mod FSteps
                else begin
                  if FCounter <= 0
                  then FCounter := FSteps - 1
                  else Dec( FCounter);
                end;
              end
              else FCounter := Random( FSteps);
              FOutValue := FInCache[ FCounter + textsequencer_step1] - MiddleNote / 128;
            end;
          end;

        seqHigh :
          begin
            if FInCache[ textsequencer_trig] <= 0
            then FInState := seqIdle;
          end;

      end;
      FOutCache[ textsequencer_out] := FOutValue + FInCache[ textsequencer_chain];
      FOldRes := SignalToLogic( FInCache[ textsequencer_reset]);
    end;

    procedure   TModTextSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInCache[ textsequencer_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInCache[ textsequencer_reset]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInCache[ textsequencer_dir  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%d' , [ aPrefix, Name, FCounter + 1                  ], AppLocale));
      aCallback( Self, Format( '%s/%s/steps/%d', [ aPrefix, Name, FSteps                        ], AppLocale));
    end;

    function    TModTextSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        textsequencer_trig  : Result := 'trig';
        textsequencer_reset : Result := 'reset';
        textsequencer_dir   : Result := 'dir';
        textsequencer_mode  : Result := 'mode';
        textsequencer_chain : Result := 'chain';
        else begin
          if anIndex - textsequencer_step1 < FSteps
          then Result := Format( 'step%d', [ anIndex - textsequencer_step1 + 1], AppLocale)
          else Result := NO_INPUT;;
        end;
      end;
    end;

    function    TModTextSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        textsequencer_out : Result := 'out';
        else                Result := NO_OUTPUT;
      end;
    end;

    procedure   TModTextSequencer.SetStepValues( const aValues: TSignals);
    begin
      FSteps := aValues.Count;
      SetInputs( textsequencer_step1, aValues);
    end;


{ ========
  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
}

const

  seqstep_triginfwd  = 0;
  seqstep_triginbwd  = 1;
  seqstep_chain      = 2;
  seqstep_reset      = 3;
  seqstep_dir        = 4;
  seqstep_note       = 5;
  seqstep_time       = 6;

  seqstep_trigoutfwd = 0;
  seqstep_trigoutbwd = 1;
  seqstep_out        = 2;


    constructor TModSeqStep.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModSeqStep.SetDefaults; // override;
    begin
      inherited;
      FInCache[ seqstep_dir] := 1; // Make seq run forward by default.
      FKnobMap := [ seqstep_time];
    end;

    procedure   TModSeqStep.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        FState := sssIdle;
        FOutCache[ seqstep_trigoutfwd] := LogicToSignal( False);                 // Do not trigger next in forward  direction
        FOutCache[ seqstep_trigoutbwd] := LogicToSignal( False);                 // Do not trigger next in backward direction
        ResetFlag := False;
      end;

      if not FOldRes and SignalToLogic( FInCache[ seqstep_reset])
      then begin                                                                 // Reset from inactive to active
        FState := sssIdle;
        FOutCache[ seqstep_trigoutfwd] := LogicToSignal( False);                 // Do not trigger next in forward  direction
        FOutCache[ seqstep_trigoutbwd] := LogicToSignal( False);                 // Do not trigger next in backward direction
      end;

      case FState of

        sssIdle :
          begin
            if SignalToLogic( FInCache[ seqstep_dir])
            then begin                                                           // running Forwards
              if SignalToLogic( FInCache[ seqstep_triginfwd])                    // Forward trig in
              then begin
                FState := sssActive;
                FTime  := SlowTimeToSampleCount( FInCache[ seqstep_time]);       // Duration
                FOutCache[ seqstep_trigoutfwd] := LogicToSignal( False);         // Do not trigger next in forward  direction
                FOutCache[ seqstep_trigoutbwd] := LogicToSignal( False);         // Do not trigger next in backward direction
              end;
            end
            else begin                                                           // running Backwards
              if SignalToLogic( FInCache[ seqstep_triginbwd])                    // Backward trig in
              then begin
                FState := sssActive;
                FTime  := SlowTimeToSampleCount( FInCache[ seqstep_time]);       // Duration
                FOutCache[ seqstep_trigoutfwd] := LogicToSignal( False);         // Do not trigger next in forward  direction
                FOutCache[ seqstep_trigoutbwd] := LogicToSignal( False);         // Do not trigger next in backward direction
              end;
            end;
          end;

        sssActive :
          begin
            FTime := FTime - 1;
            if FTime <= 0
            then begin
              FState := sssWaiting;
              if SignalToLogic( FInCache[ seqstep_dir])                          // running Forwards
              then FOutCache[ seqstep_trigoutfwd] := LogicToSignal( True)        // Trigger next in forward  direction
              else FOutCache[ seqstep_trigoutbwd] := LogicToSignal( True);       // Trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if SignalToLogic( FInCache[ seqstep_dir])
            then begin                                                           // running Forwards
              if not SignalToLogic( FInCache[ seqstep_triginfwd])                // Forward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end
            else begin                                                           // running Backwards
              if not SignalToLogic( FInCache[ seqstep_triginbwd])                // Backward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end;
          end;

      end;

      if FState = sssActive
      then FOutCache[ seqstep_out] := FInCache[ seqstep_note] - MiddleNote / 128 // Own value to output
      else FOutCache[ seqstep_out] := FInCache[ seqstep_chain];                  // Chained value to output (was compensated there already)

      FOldRes := SignalToLogic( FInCache[ seqstep_reset]);
    end;

    procedure   TModSeqStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInCache[ seqstep_triginfwd]         ], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInCache[ seqstep_triginbwd]         ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInCache[ seqstep_reset    ]         ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInCache[ seqstep_dir      ]         ], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqstep_triginfwd : Result := 'triginfwd';
        seqstep_triginbwd : Result := 'triginbwd';
        seqstep_chain     : Result := 'chain';
        seqstep_reset     : Result := 'reset';
        seqstep_dir       : Result := 'dir';
        seqstep_note      : Result := 'note';
        seqstep_time      : Result := 'time';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModSeqStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqstep_trigoutfwd : Result := 'trigoutfwd';
        seqstep_trigoutbwd : Result := 'trigoutbwd';
        seqstep_out        : Result := 'out';
        else                 Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
}

const

  seqclockstep_triginfwd  = 0;
  seqclockstep_triginbwd  = 1;
  seqclockstep_chain      = 2;
  seqclockstep_reset      = 3;
  seqclockstep_dir        = 4;
  seqclockstep_clock      = 5;
  seqclockstep_note       = 6;
  seqclockstep_counts     = 7;

  seqclockstep_trigoutfwd = 0;
  seqclockstep_trigoutbwd = 1;
  seqclockstep_out        = 2;


    constructor TModSeqClockStep.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModSeqClockStep.SetDefaults; // override;
    begin
      inherited;
      FInCache[ seqclockstep_dir] := 1; // Make seq run forward by default.
    end;

    procedure   TModSeqClockStep.DoSlowTick; // override;
    begin

      if ResetFlag
      then begin
        FState      := sssIdle;
        FTrigOutFwd := False;
        FTrigOutBwd := False;
        FOutCache[ seqclockstep_trigoutfwd] := LogicToSignal( False);                       // Do not trigger next in forward  direction
        FOutCache[ seqclockstep_trigoutbwd] := LogicToSignal( False);                       // Do not trigger next in backward direction
        FOutCache[ seqclockstep_out] := FInCache[ seqclockstep_chain];                      // Chained value to output
        ResetFlag := False;
      end;

      if not FOldRes and SignalToLogic( FInCache[ seqclockstep_reset])
      then begin                                                                            // Reset from inactive to active - reset is asynchroneous
        FState      := sssIdle;
        FTrigOutFwd := False;
        FTrigOutBwd := False;
        FOutCache[ seqclockstep_trigoutfwd] := LogicToSignal( False);                       // Do not trigger next in forward  direction
        FOutCache[ seqclockstep_trigoutbwd] := LogicToSignal( False);                       // Do not trigger next in backward direction
        FOutCache[ seqclockstep_out] := FInCache[ seqclockstep_chain];                      // Chained value to output
      end;

      FTrigInFwd := SignalToLogic( FInCache[ seqclockstep_triginfwd]);
      FTrigInBwd := SignalToLogic( FInCache[ seqclockstep_triginbwd]);
      FDir       := SignalToLogic( FInCache[ seqclockstep_dir]);

      case FState of

        sssIdle :
          begin
            if ( FDir and FTrigInFwd) or (( not FDir) and FTrigInBwd)
            then begin                                                                      // Trigger seen
              FState      := sssActive;
              FCount      := Ceil( FInCache[ seqclockstep_counts]);                         // Duration
              FTrigOutFwd := False;                                                         // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                                         // Do not trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if ( FDir and ( not FTrigInFwd)) or (( not FDir) and ( not FTrigInBwd))
            then begin
              FState      := sssIdle;                                                       // Trigger fell off - go idle
              FTrigOutFwd := False;                                                         // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                                         // Do not trigger next in backward direction
            end;
          end;
      end;

      if not FOldClock and SignalToLogic( FInCache[ seqclockstep_clock])
      then begin                                                                            // Clock from inactive to active - state changes should only occur here
        case FState of
          sssActive :
            begin
              FCount := FCount - 1;
              if FCount <= 1
              then begin
                FCount := 0;
                FState := sssWaiting;
                if FDir                                                                     // running Forwards
                then begin
                  FTrigOutFwd := True;                                                      // Trigger next in forward  direction
                  FTrigOutBwd := False;                                                     // No Trigger next in backward  direction
                end
                else begin
                  FTrigOutFwd := False;                                                     // No Trigger next in forward direction
                  FTrigOutBwd := True;                                                      // Trigger next in backward direction
                end;
              end;
            end;
        end;
      end;

      FOutCache[ seqclockstep_trigoutfwd] := LogicToSignal( FTrigOutFwd);                  // Clock out internal state to ouptputs
      FOutCache[ seqclockstep_trigoutbwd] := LogicToSignal( FTrigOutBwd);

      if FState = sssActive
      then FOutCache[ seqclockstep_out] := FInCache[ seqclockstep_note] - MiddleNote / 128  // Own     value to output
      else FOutCache[ seqclockstep_out] := FInCache[ seqclockstep_chain];                   // Chained value to output (was compensated there already)

      FOldRes   := SignalToLogic( FInCache[ seqclockstep_triginbwd]);
      FOldClock := SignalToLogic( FInCache[ seqclockstep_clock    ]);
    end;


    procedure   TModSeqClockStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInCache[ seqclockstep_triginfwd]    ], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInCache[ seqclockstep_triginbwd]    ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInCache[ seqclockstep_reset    ]    ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInCache[ seqclockstep_dir      ]    ], AppLocale));
      aCallback( Self, Format( '%s/%s/clock/%f'  , [ aPrefix, Name, FInCache[ seqclockstep_clock    ]    ], AppLocale));
      aCallback( Self, Format( '%s/%s/count/%d'  , [ aPrefix, Name, FCount                               ], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqClockStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqclockstep_triginfwd : Result := 'triginfwd';
        seqclockstep_triginbwd : Result := 'triginbwd';
        seqclockstep_chain     : Result := 'chain';
        seqclockstep_reset     : Result := 'reset';
        seqclockstep_dir       : Result := 'dir';
        seqclockstep_clock     : Result := 'clock';
        seqclockstep_note      : Result := 'note';
        seqclockstep_counts    : Result := 'counts';
        else                     Result := NO_INPUT;
      end;
    end;

    function    TModSeqClockStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqclockstep_trigoutfwd : Result := 'trigoutfwd';
        seqclockstep_trigoutbwd : Result := 'trigoutbwd';
        seqclockstep_out        : Result := 'out';
        else                      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMemory = class( TMod)
  private
    FMemoryL : TSignals;
    FMemoryR : TSignals;
  public
}

    constructor TModMemory.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FMemoryL := TSignals.Create( 0);
      FMemoryR := TSignals.Create( 0);
    end;

    destructor  TModMemory.Destroy; // override;
    begin
      FreeAndNil( FMemoryL);
      FreeAndNil( FMemoryR);
      inherited;
    end;

    function    TModMemory.Convolve( aSample: TSignal): TSignalPair;
    begin
      Result.Left  := FMemoryL.Convolve( aSample);
      Result.Right := FMemoryR.Convolve( aSample);
    end;


{ ========
  TModDelay = class( TModMemory)
  private
    FLength    : TSignal;
    FInPointer : Integer;
  public
}

const

  delay_in       = 0;
  delay_delay    = 1;

  delay_out      = 0;
  delay_outfixed = 1;


    procedure   TModDelay.SetDefaults; // override;
    begin
      inherited;
      FInCache[ delay_delay] := 1.0; // Set maximum delay time if nothing connected
    end;

    procedure   TModDelay.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      aMemory: TSignals;
      bMemory: TSignals;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 0.000001
        then aValue := 0.000001;
        if aValue <> FMemLength
        then begin
          aMemory    := TSignals.Create( Ceil( TimeToSampleCount( aValue)));
          bMemory    := FMemoryL;
          FMemoryL   := aMemory;
          FMemLength := aValue;

          FreeAndNil( bMemory);
        end;
      end
      else inherited;
    end;

    function    TModDelay.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FMemLength
      else Result := inherited;
    end;

    procedure   TModDelay.DoTick; // override;
    var
      OutPointerVar : TSignal;
      SampCount     : Integer;
    begin
      SampCount                    := FMemoryL.Count;
      FInPointer                   := FInPointer mod SampCount;
      FOutCache[ delay_outfixed]   := FMemoryL.Signal[ FInPointer];
      FMemoryL.Signal[ FInPointer] := Clip( FInCache[ delay_in], -16, 16);
      OutPointerVar                := FModulo( SampCount + FInPointer - Clip( FInCache[ delay_delay], 0.0, 1.0) * ( TimeToSampleCount( FMemLength) - 1), SampCount);
      FOutCache[ delay_out]        := AntiRunAway( FMemoryL.Interpolate( OutPointerVar));
      FInPointer                   := ( FInPointer + 1) mod SampCount;
    end;

    function    TModDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        delay_in    : Result := 'in';
        delay_delay : Result := 'delay';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        delay_out      : Result := 'out';
        delay_outfixed : Result := 'outfixed';
        else             Result := NO_OUTPUT;
      end;
    end;

    function    TModDelay.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'length';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModWave = class( TModMemory)
  private
    FFileName    : string;
    FDuration    : string;
    FIsStereo    : Boolean;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

const

  wave_outl = 0;
  wave_outr = 1;


    procedure   TModWave.SetDuration( aValue: Cardinal);
    begin
      if aValue < 1000
      then FDuration := Format( '%.3f ms', [ aValue / 1.0], AppLocale)
      else if aValue >= 10000
      then FDuration := Format( '%.2d:%.2d', [ Round( aValue / 1000) div 60, Round( aValue / 1000) mod 60])
      else FDuration := Format( '%.3f s', [ aValue / 1000.0], AppLocale);
    end;

    procedure   TModWave.SwapMemoryFor( aMemory: TSignalArray);
    var
      aNewMemoryL  : TSignals;
      anOldMemoryL : TSignals;
      aNewMemoryR  : TSignals;
      anOldMemoryR : TSignals;
      i            : Integer;
    begin
      if FIsStereo
      then begin
        aNewMemoryL  := TSignals.Create( Length( aMemory) div 2);
        aNewMemoryR  := TSignals.Create( Length( aMemory) div 2);
        for i := 0 to aNewMemoryL.Count - 1
        do begin
          aNewMemoryL.Signal[ i] := aMemory[ 2 * i    ];
          aNewMemoryR.Signal[ i] := aMemory[ 2 * i + 1];
        end;

        anOldMemoryL := FMemoryL;
        anOldMemoryR := FMemoryR;
        FMemoryL     := aNewMemoryL;
        FMemoryR     := aNewMemoryR;

        FreeAndNil( anOldMemoryL);
        FreeAndNil( anOldMemoryR);
      end
      else begin
        aNewMemoryL  := TSignals.Create( Length( aMemory));
        for i := 0 to Length( aMemory) - 1
        do aNewMemoryL.Signal[ i] := aMemory[ i];

        anOldMemoryL := FMemoryL;
        FMemoryL     := aNewMemoryL;

        FreeAndNil( anOldMemoryL);
      end;
    end;

    procedure   TModWave.LoadFile;
    const
      BufSize = 8192;
    type
      TSample = SmallInt;
      TBuffer = array[ 0 .. BufSize - 1] of TSample;

      procedure SetSamples16bit44100Hz( var aMemory: TSignalArray; aBuf: TBuffer; aSampleCount: Cardinal);
      const
        MaxVal = High( TSample) / 2.0;
      var
        i        : Cardinal;
        Position : Cardinal;
      begin
        Position := Length( aMemory);
        SetLength( aMemory, Position + aSampleCount);
        for i := 0 to aSampleCount - 1
        do begin
          aMemory[ Position] := aBuf[ i] / MaxVal;
          Inc( Position);
        end;
      end;

    var
      aMemory    : TSignalArray;
      aWaveFile  : TWaveFile;
      aByteCount : Cardinal;
      aBuf       : TBuffer;
      aDuration  : Cardinal;
      aFormat    : TPCMFormat;
    begin
      SetLength( aMemory, 0);
      aDuration := 0;
      if FileExists( FileName) // todo: and FileSize < 44100 * 2 * 2 * 301 // 44k1, stereo, 16 bit and duration < 301 s.
      then begin
        aWaveFile := TWaveFile.Create( FileName, fmOpenRead);
        try
          if aWaveFile.Valid
          then begin
            aFormat := aWaveFile.PCMFormat;
            if ( aFormat = Stereo16bit44100Hz) or ( aFormat = Mono16bit44100Hz)
            then begin
              aDuration := aWaveFile.Length;
              aWaveFile.Position := 0;
              if aWaveFile.BeginRead
              then begin
                try
                  while Cardinal( aWaveFile.Position) < aWaveFile.DataSize
                  do begin
                    aByteCount := aWaveFile.Read( aBuf, BufSize * SizeOf( TSample));
                    case aFormat of
                      Stereo16bit44100Hz : begin FIsStereo := True ; SetSamples16bit44100Hz( aMemory, aBuf, aByteCount div SizeOf( TSample)); end;
                      Mono16bit44100Hz   : begin FIsStereo := False; SetSamples16bit44100Hz( aMemory, aBuf, aByteCount div SizeOf( TSample)); end;
                    end;
                  end;
                finally
                  aWaveFile.EndRead;
                end;
              end;
            end;
          end;
        finally
          aWaveFile.Free;
        end;
      end;
      SwapMemoryFor( aMemory);
      SetDuration( aDuration);
    end;

    procedure   TModWave.SetFileName( const aValue: string);
    begin
      if  aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

//  public

    destructor  TModWave.Destroy; // override;
    begin
      inherited;
    end;

    procedure   TModWave.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModWave.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModWave.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    const
      MonoText : array[ Boolean] of string = ( 'M', 'S');
    begin
      aCallback( Self, Format( '%s/%s/duration/%s', [ aPrefix, Name, FDuration]           , AppLocale));
      aCallback( Self, Format( '%s/%s/mono/%s'    , [ aPrefix, Name, MonoText[ FIsStereo]], AppLocale));
    end;

    function    TModWave.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        wave_outl :  Result := 'out';        // Left  output
        wave_outr :  Result := 'outr';       // Right output
        else         Result := NO_OUTPUT;
      end;
    end;

    function    TModWave.StringName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'filename';
        else Result := NO_STRING;
      end;
    end;



{ ========
  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
}

const

  waveplayer_speed     = 0;
  waveplayer_position  = 1;
  waveplayer_res       = 2;
  waveplayer_mode      = 3;
  waveplayer_fm        = 4;
{ waveplayer_convolve  = 5; }


    procedure   TModWavePlayer.SetDefaults; // override;
    begin
      inherited;
      FInCache[ waveplayer_speed]    := 0.0; // Set zero speed
      FInCache[ waveplayer_position] := 0.0; // Set start position
      FInCache[ waveplayer_res]      := 0.0; // Reset to zero
      FInCache[ waveplayer_fm]       := 0.0; // FrequencyToUnits( ReferenceA); // Set FM amount to neutral to make speed work ...
      SetDuration( 0);
    end;

    procedure   TModWavePlayer.DoTick; // override;
    var
      SampCount : Integer;
      Increment : TSignal;
      Position  : TSignal;
      MakeMono  : Boolean;
    begin
      if ResetFlag
      then begin
        FPlayPointer := 0.0;
        ResetFlag    := False;
      end;

      if ( not FOldRes) and SignalToLogic( FInCache[ waveplayer_res])
      then FPlayPointer := 0.0;
      FOldRes := SignalToLogic( FInCache[ waveplayer_res]);

      if FIsStereo
      then begin
        MakeMono := SignalToInt( FInCache[ waveplayer_mode]) = 0;
        if Assigned( FMemoryL) and Assigned( FMemoryR)
        then begin
          SampCount := FMemoryL.Count;
          if SampCount > 0
          then begin
            Increment    := Clip( FInCache[ waveplayer_speed] * LookupFrequency( FInCache[ waveplayer_fm]) / LookupFrequency( 0), - 16, 16);
            Position     := ( 32 + Clip( FInCache[ waveplayer_position],  -16, 16)) * SampCount;
            FPlayPointer := FModulo( FPlayPointer + Increment + SampCount, SampCount);
            Position     := FModulo( Position + FPlayPointer, SampCount);
            if MakeMono
            then begin
              FOutCache[ wave_outl] := ( FMemoryL.Interpolate( Position) + FMemoryR.Interpolate( Position)) / 2;
              FOutCache[ wave_outr] := FOutCache[ wave_outl];
            end
            else begin
              FOutCache[ wave_outl] := FMemoryL.Interpolate( Position);
              FOutCache[ wave_outr] := FMemoryR.Interpolate( Position);
            end;
          end
          else begin
            FOutCache[ wave_outl] := 0;
            FOutCache[ wave_outr] := 0;
          end;
        end
        else begin
          FOutCache[ wave_outl] := 0;
          FOutCache[ wave_outr] := 0;
        end;
      end
      else begin
        if Assigned( FMemoryL)
        then begin
          SampCount := FMemoryL.Count;
          if SampCount > 0
          then begin
            Increment     := Clip( FInCache[ waveplayer_speed] * LookupFrequency( FInCache[ waveplayer_fm]) / LookupFrequency( 0), - 16, 16);
            Position      := ( 32 + Clip( FInCache[ waveplayer_position],  -16, 16)) * SampCount;
            FPlayPointer  := FModulo( FPlayPointer + Increment + SampCount, SampCount);
            Position      := FModulo( Position + FPlayPointer, SampCount);
            FOutCache[ wave_outl] := FMemoryL.Interpolate( Position);
          end
          else FOutCache[ wave_outl] := 0;
        end
        else FOutCache[ wave_outl] := 0;
        FOutCache[ wave_outr] := FOutCache[ wave_outl];
      end;
    end;

    procedure   TModWavePlayer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f', [ aPrefix, Name, FInCache[ waveplayer_res]], AppLocale));
    end;

    function    TModWavePlayer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        waveplayer_speed    : Result := 'speed';
        waveplayer_position : Result := 'position';
        waveplayer_res      : Result := 'res';
        waveplayer_mode     : Result := 'mode';            // 0 = mono, 1 = stereo
        waveplayer_fm       : Result := 'fm';
     // waveplayer_convolve : Result := 'convolve';
        else                  Result := NO_INPUT;
      end;
    end;



{ ========
  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
}

const

  sampleandhold_trig = 0;
  sampleandhold_in   = 1;

  sampleandhold_out  = 0;


    procedure   TModSampleAndHold.DoTick; // override;
    var
      Trig : Boolean;
    begin
      Trig := SignalToLogic( FInCache[ sampleandhold_trig]);
      if not FLastTrig and Trig
      then FOutCache[ sampleandhold_out] := FInCache[ sampleandhold_in];
      FLastTrig := Trig;
    end;

    procedure   TModSampleAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ sampleandhold_trig]], AppLocale));
    end;

    function    TModSampleAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sampleandhold_trig : Result := 'trig';
        sampleandhold_in   : Result := 'in';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModSampleAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sampleandhold_out : Result := 'out';
        else                Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrackAndHold = class( TMod)
  public
}

const

  trackandhold_trig = 0;
  trackandhold_in   = 1;

  trackandhold_out  = 0;


    procedure   TModTrackAndHold.DoTick; // override;
    begin
      if SignalToLogic( FInCache[ trackandhold_trig])
      then FOutCache[ trackandhold_out] := FInCache[ trackandhold_in];
    end;

    procedure   TModTrackAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ trackandhold_trig]], AppLocale));
    end;

    function    TModTrackAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        trackandhold_trig : Result := 'trig';
        trackandhold_in   : Result := 'in';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModTrackAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        trackandhold_out : Result := 'out';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSwitch2to1 = class( TMod)
  public
}

const

    switch2to1_control = 0;
    switch2to1_in1     = 1;
    switch2to1_in2     = 2;

    switch2to1_out     = 0;


    procedure   TModSwitch2to1.DoTick; // override;
    begin
      if SignalToLogic( FInCache[ switch2to1_control])
      then FOutCache[ switch2to1_out] := FInCache[ switch2to1_in2]
      else FOutCache[ switch2to1_out] := FInCache[ switch2to1_in1];
    end;

    procedure   TModSwitch2to1.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/control/%f', [ aPrefix, Name, FInCache[ switch2to1_control]], AppLocale));
    end;

    function    TModSwitch2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        switch2to1_control : Result := 'control';
        switch2to1_in1     : Result := 'in1';
        switch2to1_in2     : Result := 'in2';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModSwitch2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        switch2to1_out : Result := 'out';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModXFade = class( TMod)
  public
}

const

  xfade_control  = 0;
  xfade_in1      = 1;
  xfade_in2      = 2;
  xfade_position = 3;

  xfade_out      = 0;


    procedure   TModXFade.SetDefaults; // override;
    begin
      FInCache[ xfade_control ] := 0; // Equal amounts of in1 and in2
      FInCache[ xfade_position] := 0;
      FKnobMap := [ xfade_position];
    end;

    procedure   TModXFade.DoTick; // override;
    var
      Control: TSignal;
    begin
      Control    := FInCache[ xfade_control] + FInCache[ xfade_position];
      FOutCache[ xfade_out] := AntiRunAway((( 1 - Control) * FInCache[ xfade_in1] + ( 1 + Control) * FInCache[ xfade_in2]) / 2);
    end;

    function    TModXFade.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        xfade_control  : Result := 'control';
        xfade_in1      : Result := 'in1';
        xfade_in2      : Result := 'in2';
        xfade_position : Result := 'position';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModXFade.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        xfade_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModPan = class( TMod)
  public
}

const

  pan_control  = 0;
  pan_in       = 1;
  pan_position = 2;

  pan_out1     = 0;
  pan_out2     = 1;


    procedure   TModPan.SetDefaults; // override;
    begin
      FInCache[ pan_control ] := 0; // Mid position
      FKnobMap := [ pan_position];
    end;

    procedure   TModPan.DoTick; // override;
    var
      Control : TSignal;
      Inp     : TSignal;
    begin
      Control    := FInCache[ pan_control] + FInCache[ pan_position];
      Inp        := FInCache[ pan_in];
      FOutCache[ pan_out1] := AntiRunAway((( 1 - Control) * Inp) / 2);
      FOutCache[ pan_out2] := AntiRunAway((( 1 + Control) * Inp) / 2);
    end;

    function    TModPan.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pan_control  : Result := 'control';
        pan_in       : Result := 'in';
        pan_position : Result := 'position';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModPan.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pan_out1 : Result := 'out1';
        pan_out2 : Result := 'out2';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixM4to1 = class( TMod)
  public
}

const

  mixm4to1_in     =  0;
  mixm4to1_in1    =  1;
  mixm4to1_in2    =  2;
  mixm4to1_in3    =  3;
  mixm4to1_in4    =  4;
  mixm4to1_mute1  =  5;
  mixm4to1_mute2  =  6;
  mixm4to1_mute3  =  7;
  mixm4to1_mute4  =  8;
  mixm4to1_level1 =  9;
  mixm4to1_level2 = 10;
  mixm4to1_level3 = 11;
  mixm4to1_level4 = 12;

  mixm4to1_out    =  0;


    procedure   TModMixM4to1.SetDefaults; // override;
    begin
      FInCache[ mixm4to1_in ] := 0; // Unused inputs should be zero
      FInCache[ mixm4to1_in1] := 0;
      FInCache[ mixm4to1_in2] := 0;
      FInCache[ mixm4to1_in3] := 0;
      FInCache[ mixm4to1_in4] := 0;
      FKnobMap := [
        mixm4to1_mute1,
        mixm4to1_mute2,
        mixm4to1_mute3,
        mixm4to1_mute4,
        mixm4to1_level1,
        mixm4to1_level2,
        mixm4to1_level3,
        mixm4to1_level4
      ];
    end;

    procedure   TModMixM4to1.DoTick; // override;
    begin
      FOutCache[ mixm4to1_out] :=
        FInCache[ mixm4to1_in ] +
        FInCache[ mixm4to1_in1] * FInCache[ mixm4to1_level1] * SignalToMute( FInCache[ mixm4to1_mute1]) +
        FInCache[ mixm4to1_in2] * FInCache[ mixm4to1_level2] * SignalToMute( FInCache[ mixm4to1_mute2]) +
        FInCache[ mixm4to1_in3] * FInCache[ mixm4to1_level3] * SignalToMute( FInCache[ mixm4to1_mute3]) +
        FInCache[ mixm4to1_in4] * FInCache[ mixm4to1_level4] * SignalToMute( FInCache[ mixm4to1_mute4]) ;
    end;

    function    TModMixM4to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm4to1_in     : Result := 'in';
        mixm4to1_in1    : Result := 'in1';
        mixm4to1_in2    : Result := 'in2';
        mixm4to1_in3    : Result := 'in3';
        mixm4to1_in4    : Result := 'in4';
        mixm4to1_mute1  : Result := 'mute1';
        mixm4to1_mute2  : Result := 'mute2';
        mixm4to1_mute3  : Result := 'mute3';
        mixm4to1_mute4  : Result := 'mute4';
        mixm4to1_level1 : Result := 'level1';
        mixm4to1_level2 : Result := 'level2';
        mixm4to1_level3 : Result := 'level3';
        mixm4to1_level4 : Result := 'level4';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModMixM4to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm4to1_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixS2to1 = class( TMod)
  public
}

const

  mixs2to1_inl    = 0;
  mixs2to1_inr    = 1;
  mixs2to1_inl1   = 2;
  mixs2to1_inr1   = 3;
  mixs2to1_inl2   = 4;
  mixs2to1_inr2   = 5;
  mixs2to1_mute1  = 6;
  mixs2to1_mute2  = 7;
  mixs2to1_level1 = 8;
  mixs2to1_level2 = 9;

  mixs2to1_outl   = 0;
  mixs2to1_outr   = 1;


    procedure   TModMixS2to1.SetDefaults; // override;
    begin
      FInCache[ mixs2to1_inl ] := 0; // Unused inputs should be zero
      FInCache[ mixs2to1_inr ] := 0;
      FInCache[ mixs2to1_inl1] := 0;
      FInCache[ mixs2to1_inr1] := 0;
      FInCache[ mixs2to1_inl2] := 0;
      FInCache[ mixs2to1_inr2] := 0;
      FKnobMap := [
        mixs2to1_mute1,
        mixs2to1_mute2,
        mixs2to1_level1,
        mixs2to1_level2
      ];
    end;

    procedure   TModMixS2to1.DoTick; // override;
    var
      lev1 : TSignal;
      lev2 : TSignal;
    begin
      lev1 := FInCache[ mixs2to1_level1] * SignalToMute( FInCache[ mixs2to1_mute1]);
      lev2 := FInCache[ mixs2to1_level2] * SignalToMute( FInCache[ mixs2to1_mute2]);

      FOutCache[ mixs2to1_outl] :=
        FInCache[ mixs2to1_inl]         +
        FInCache[ mixs2to1_inl1] * lev1 +
        FInCache[ mixs2to1_inl2] * lev2 ;
      FOutCache[ mixs2to1_outr] :=
        FInCache[ mixs2to1_inr]         +
        FInCache[ mixs2to1_inr1] * lev1 +
        FInCache[ mixs2to1_inr2] * lev2 ;
    end;

    function    TModMixS2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixs2to1_inl    : Result := 'inl';
        mixs2to1_inr    : Result := 'inr';
        mixs2to1_inl1   : Result := 'inl1';
        mixs2to1_inr1   : Result := 'inr1';
        mixs2to1_inl2   : Result := 'inl2';
        mixs2to1_inr2   : Result := 'inr2';
        mixs2to1_mute1  : Result := 'mute1';
        mixs2to1_mute2  : Result := 'mute2';
        mixs2to1_level1 : Result := 'level1';
        mixs2to1_level2 : Result := 'level2';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModMixS2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixs2to1_outl : Result := 'outl';
        mixs2to1_outr : Result := 'outr';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
 TModMixM6to6 = class( TMod)
  public
}

const

  mixm6to6_in1       =  0;
  mixm6to6_in2       =  1;
  mixm6to6_in3       =  2;
  mixm6to6_in4       =  3;
  mixm6to6_in5       =  4;
  mixm6to6_in6       =  5;
  mixm6to6_mutel     =  6;
  mixm6to6_mute2     =  7;
  mixm6to6_mute3     =  8;
  mixm6to6_mute4     =  9;
  mixm6to6_mute5     = 10;
  mixm6to6_mute6     = 11;
  mixm6to6_inchcol1  = 12;
  mixm6to6_inchcol2  = 13;
  mixm6to6_inchcol3  = 14;
  mixm6to6_inchcol4  = 15;
  mixm6to6_inchcol5  = 16;
  mixm6to6_inchcol6  = 17;
  mixm6to6_levelr1c1 = 18;
  mixm6to6_levelr1c2 = 19;
  mixm6to6_levelr1c3 = 20;
  mixm6to6_levelr1c4 = 21;
  mixm6to6_levelr1c5 = 22;
  mixm6to6_levelr1c6 = 23;
  mixm6to6_levelr2c1 = 24;
  mixm6to6_levelr2c2 = 25;
  mixm6to6_levelr2c3 = 26;
  mixm6to6_levelr2c4 = 27;
  mixm6to6_levelr2c5 = 28;
  mixm6to6_levelr2c6 = 29;
  mixm6to6_levelr3c1 = 30;
  mixm6to6_levelr3c2 = 31;
  mixm6to6_levelr3c3 = 32;
  mixm6to6_levelr3c4 = 33;
  mixm6to6_levelr3c5 = 34;
  mixm6to6_levelr3c6 = 35;
  mixm6to6_levelr4c1 = 36;
  mixm6to6_levelr4c2 = 37;
  mixm6to6_levelr4c3 = 38;
  mixm6to6_levelr4c4 = 39;
  mixm6to6_levelr4c5 = 40;
  mixm6to6_levelr4c6 = 41;
  mixm6to6_levelr5c1 = 42;
  mixm6to6_levelr5c2 = 43;
  mixm6to6_levelr5c3 = 44;
  mixm6to6_levelr5c4 = 45;
  mixm6to6_levelr5c5 = 46;
  mixm6to6_levelr5c6 = 47;
  mixm6to6_levelr6c1 = 48;
  mixm6to6_levelr6c2 = 49;
  mixm6to6_levelr6c3 = 50;
  mixm6to6_levelr6c4 = 51;
  mixm6to6_levelr6c5 = 52;
  mixm6to6_levelr6c6 = 53;
  mixm6to6_inr1c1    = 54;
  mixm6to6_inr1c2    = 55;
  mixm6to6_inr1c3    = 56;
  mixm6to6_inr1c4    = 57;
  mixm6to6_inr1c5    = 58;
  mixm6to6_inr1c6    = 59;
  mixm6to6_inr2c1    = 60;
  mixm6to6_inr2c2    = 61;
  mixm6to6_inr2c3    = 62;
  mixm6to6_inr2c4    = 63;
  mixm6to6_inr2c5    = 64;
  mixm6to6_inr2c6    = 65;
  mixm6to6_inr3c1    = 66;
  mixm6to6_inr3c2    = 67;
  mixm6to6_inr3c3    = 68;
  mixm6to6_inr3c4    = 69;
  mixm6to6_inr3c5    = 70;
  mixm6to6_inr3c6    = 71;
  mixm6to6_inr4c1    = 72;
  mixm6to6_inr4c2    = 73;
  mixm6to6_inr4c3    = 74;
  mixm6to6_inr4c4    = 75;
  mixm6to6_inr4c5    = 76;
  mixm6to6_inr4c6    = 77;
  mixm6to6_inr5c1    = 78;
  mixm6to6_inr5c2    = 79;
  mixm6to6_inr5c3    = 80;
  mixm6to6_inr5c4    = 81;
  mixm6to6_inr5c5    = 82;
  mixm6to6_inr5c6    = 83;
  mixm6to6_inr6c1    = 84;
  mixm6to6_inr6c2    = 85;
  mixm6to6_inr6c3    = 86;
  mixm6to6_inr6c4    = 87;
  mixm6to6_inr6c5    = 88;
  mixm6to6_inr6c6    = 89;

  mixm6to6_out1      =  0;
  mixm6to6_out2      =  1;
  mixm6to6_out3      =  2;
  mixm6to6_out4      =  3;
  mixm6to6_out5      =  4;
  mixm6to6_out6      =  5;
  mixm6to6_outchrow1 =  6;
  mixm6to6_outchrow2 =  7;
  mixm6to6_outchrow3 =  8;
  mixm6to6_outchrow4 =  9;
  mixm6to6_outchrow5 = 10;
  mixm6to6_outchrow6 = 11;


    procedure   TModMixM6to6.SetDefaults; // override;
    begin
      FInCache[ mixm6to6_inchcol1] := 0.0; // Zero chain col inputs
      FInCache[ mixm6to6_inchcol2] := 0.0;
      FInCache[ mixm6to6_inchcol3] := 0.0;
      FInCache[ mixm6to6_inchcol4] := 0.0;
      FInCache[ mixm6to6_inchcol5] := 0.0;
      FInCache[ mixm6to6_inchcol6] := 0.0;
      FKnobMap := [ mixm6to6_mutel .. mixm6to6_mute6, mixm6to6_levelr1c1 .. mixm6to6_levelr6c6];
    end;

    procedure   TModMixM6to6.DoTick; // override;
    var
      In1 : TSignal;
      In2 : TSignal;
      In3 : TSignal;
      In4 : TSignal;
      In5 : TSignal;
      In6 : TSignal;
    begin
      In1 := FInCache[ mixm6to6_in1] * SignalToMute( FInCache[ mixm6to6_mutel]);
      In2 := FInCache[ mixm6to6_in2] * SignalToMute( FInCache[ mixm6to6_mute2]);
      In3 := FInCache[ mixm6to6_in3] * SignalToMute( FInCache[ mixm6to6_mute3]);
      In4 := FInCache[ mixm6to6_in4] * SignalToMute( FInCache[ mixm6to6_mute4]);
      In5 := FInCache[ mixm6to6_in5] * SignalToMute( FInCache[ mixm6to6_mute5]);
      In6 := FInCache[ mixm6to6_in6] * SignalToMute( FInCache[ mixm6to6_mute6]);

      FOutCache[ mixm6to6_out1] := In1 * ( FInCache[ mixm6to6_levelr1c1] + FInCache[ mixm6to6_inr1c1]) + In2 * ( FInCache[ mixm6to6_levelr2c1] + FInCache[ mixm6to6_inr2c1]) + In3 * ( FInCache[ mixm6to6_levelr3c1] + FInCache[ mixm6to6_inr3c1]) + In4 * ( FInCache[ mixm6to6_levelr4c1] + FInCache[ mixm6to6_inr4c1]) + In5 * ( FInCache[ mixm6to6_levelr5c1] + FInCache[ mixm6to6_inr5c1]) + In6 * ( FInCache[ mixm6to6_levelr6c1] + FInCache[ mixm6to6_inr6c1]) + FInCache[ mixm6to6_inchcol1];
      FOutCache[ mixm6to6_out2] := In1 * ( FInCache[ mixm6to6_levelr1c2] + FInCache[ mixm6to6_inr1c2]) + In2 * ( FInCache[ mixm6to6_levelr2c2] + FInCache[ mixm6to6_inr2c2]) + In3 * ( FInCache[ mixm6to6_levelr3c2] + FInCache[ mixm6to6_inr3c2]) + In4 * ( FInCache[ mixm6to6_levelr4c2] + FInCache[ mixm6to6_inr4c2]) + In5 * ( FInCache[ mixm6to6_levelr5c2] + FInCache[ mixm6to6_inr5c2]) + In6 * ( FInCache[ mixm6to6_levelr6c2] + FInCache[ mixm6to6_inr6c2]) + FInCache[ mixm6to6_inchcol2];
      FOutCache[ mixm6to6_out3] := In1 * ( FInCache[ mixm6to6_levelr1c3] + FInCache[ mixm6to6_inr1c3]) + In2 * ( FInCache[ mixm6to6_levelr2c3] + FInCache[ mixm6to6_inr2c3]) + In3 * ( FInCache[ mixm6to6_levelr3c3] + FInCache[ mixm6to6_inr3c3]) + In4 * ( FInCache[ mixm6to6_levelr4c3] + FInCache[ mixm6to6_inr4c3]) + In5 * ( FInCache[ mixm6to6_levelr5c3] + FInCache[ mixm6to6_inr5c3]) + In6 * ( FInCache[ mixm6to6_levelr6c3] + FInCache[ mixm6to6_inr6c3]) + FInCache[ mixm6to6_inchcol3];
      FOutCache[ mixm6to6_out4] := In1 * ( FInCache[ mixm6to6_levelr1c4] + FInCache[ mixm6to6_inr1c4]) + In2 * ( FInCache[ mixm6to6_levelr2c4] + FInCache[ mixm6to6_inr2c4]) + In3 * ( FInCache[ mixm6to6_levelr3c4] + FInCache[ mixm6to6_inr3c4]) + In4 * ( FInCache[ mixm6to6_levelr4c4] + FInCache[ mixm6to6_inr4c4]) + In5 * ( FInCache[ mixm6to6_levelr5c4] + FInCache[ mixm6to6_inr5c4]) + In6 * ( FInCache[ mixm6to6_levelr6c4] + FInCache[ mixm6to6_inr6c4]) + FInCache[ mixm6to6_inchcol4];
      FOutCache[ mixm6to6_out5] := In1 * ( FInCache[ mixm6to6_levelr1c5] + FInCache[ mixm6to6_inr1c5]) + In2 * ( FInCache[ mixm6to6_levelr2c5] + FInCache[ mixm6to6_inr2c5]) + In3 * ( FInCache[ mixm6to6_levelr3c5] + FInCache[ mixm6to6_inr3c5]) + In4 * ( FInCache[ mixm6to6_levelr4c5] + FInCache[ mixm6to6_inr4c5]) + In5 * ( FInCache[ mixm6to6_levelr5c5] + FInCache[ mixm6to6_inr5c5]) + In6 * ( FInCache[ mixm6to6_levelr6c5] + FInCache[ mixm6to6_inr6c5]) + FInCache[ mixm6to6_inchcol5];
      FOutCache[ mixm6to6_out6] := In1 * ( FInCache[ mixm6to6_levelr1c6] + FInCache[ mixm6to6_inr1c6]) + In2 * ( FInCache[ mixm6to6_levelr2c6] + FInCache[ mixm6to6_inr2c6]) + In3 * ( FInCache[ mixm6to6_levelr3c6] + FInCache[ mixm6to6_inr3c6]) + In4 * ( FInCache[ mixm6to6_levelr4c6] + FInCache[ mixm6to6_inr4c6]) + In5 * ( FInCache[ mixm6to6_levelr5c6] + FInCache[ mixm6to6_inr5c6]) + In6 * ( FInCache[ mixm6to6_levelr6c6] + FInCache[ mixm6to6_inr6c6]) + FInCache[ mixm6to6_inchcol6];

      FOutCache[ mixm6to6_outchrow1] := In1;
      FOutCache[ mixm6to6_outchrow2] := In2;
      FOutCache[ mixm6to6_outchrow3] := In3;
      FOutCache[ mixm6to6_outchrow4] := In4;
      FOutCache[ mixm6to6_outchrow5] := In5;
      FOutCache[ mixm6to6_outchrow6] := In6;
    end;

    function    TModMixM6to6.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm6to6_in1       : Result := 'in1';
        mixm6to6_in2       : Result := 'in2';
        mixm6to6_in3       : Result := 'in3';
        mixm6to6_in4       : Result := 'in4';
        mixm6to6_in5       : Result := 'in5';
        mixm6to6_in6       : Result := 'in6';
        mixm6to6_mutel     : Result := 'mutel';
        mixm6to6_mute2     : Result := 'mute2';
        mixm6to6_mute3     : Result := 'mute3';
        mixm6to6_mute4     : Result := 'mute4';
        mixm6to6_mute5     : Result := 'mute5';
        mixm6to6_mute6     : Result := 'mute6';
        mixm6to6_inchcol1  : Result := 'inchcol1';
        mixm6to6_inchcol2  : Result := 'inchcol2';
        mixm6to6_inchcol3  : Result := 'inchcol3';
        mixm6to6_inchcol4  : Result := 'inchcol4';
        mixm6to6_inchcol5  : Result := 'inchcol5';
        mixm6to6_inchcol6  : Result := 'inchcol6';
        mixm6to6_levelr1c1 : Result := 'levelr1c1';
        mixm6to6_levelr1c2 : Result := 'levelr1c2';
        mixm6to6_levelr1c3 : Result := 'levelr1c3';
        mixm6to6_levelr1c4 : Result := 'levelr1c4';
        mixm6to6_levelr1c5 : Result := 'levelr1c5';
        mixm6to6_levelr1c6 : Result := 'levelr1c6';
        mixm6to6_levelr2c1 : Result := 'levelr2c1';
        mixm6to6_levelr2c2 : Result := 'levelr2c2';
        mixm6to6_levelr2c3 : Result := 'levelr2c3';
        mixm6to6_levelr2c4 : Result := 'levelr2c4';
        mixm6to6_levelr2c5 : Result := 'levelr2c5';
        mixm6to6_levelr2c6 : Result := 'levelr2c6';
        mixm6to6_levelr3c1 : Result := 'levelr3c1';
        mixm6to6_levelr3c2 : Result := 'levelr3c2';
        mixm6to6_levelr3c3 : Result := 'levelr3c3';
        mixm6to6_levelr3c4 : Result := 'levelr3c4';
        mixm6to6_levelr3c5 : Result := 'levelr3c5';
        mixm6to6_levelr3c6 : Result := 'levelr3c6';
        mixm6to6_levelr4c1 : Result := 'levelr4c1';
        mixm6to6_levelr4c2 : Result := 'levelr4c2';
        mixm6to6_levelr4c3 : Result := 'levelr4c3';
        mixm6to6_levelr4c4 : Result := 'levelr4c4';
        mixm6to6_levelr4c5 : Result := 'levelr4c5';
        mixm6to6_levelr4c6 : Result := 'levelr4c6';
        mixm6to6_levelr5c1 : Result := 'levelr5c1';
        mixm6to6_levelr5c2 : Result := 'levelr5c2';
        mixm6to6_levelr5c3 : Result := 'levelr5c3';
        mixm6to6_levelr5c4 : Result := 'levelr5c4';
        mixm6to6_levelr5c5 : Result := 'levelr5c5';
        mixm6to6_levelr5c6 : Result := 'levelr5c6';
        mixm6to6_levelr6c1 : Result := 'levelr6c1';
        mixm6to6_levelr6c2 : Result := 'levelr6c2';
        mixm6to6_levelr6c3 : Result := 'levelr6c3';
        mixm6to6_levelr6c4 : Result := 'levelr6c4';
        mixm6to6_levelr6c5 : Result := 'levelr6c5';
        mixm6to6_levelr6c6 : Result := 'levelr6c6';
        mixm6to6_inr1c1    : Result := 'inr1c1';
        mixm6to6_inr1c2    : Result := 'inr1c2';
        mixm6to6_inr1c3    : Result := 'inr1c3';
        mixm6to6_inr1c4    : Result := 'inr1c4';
        mixm6to6_inr1c5    : Result := 'inr1c5';
        mixm6to6_inr1c6    : Result := 'inr1c6';
        mixm6to6_inr2c1    : Result := 'inr2c1';
        mixm6to6_inr2c2    : Result := 'inr2c2';
        mixm6to6_inr2c3    : Result := 'inr2c3';
        mixm6to6_inr2c4    : Result := 'inr2c4';
        mixm6to6_inr2c5    : Result := 'inr2c5';
        mixm6to6_inr2c6    : Result := 'inr2c6';
        mixm6to6_inr3c1    : Result := 'inr3c1';
        mixm6to6_inr3c2    : Result := 'inr3c2';
        mixm6to6_inr3c3    : Result := 'inr3c3';
        mixm6to6_inr3c4    : Result := 'inr3c4';
        mixm6to6_inr3c5    : Result := 'inr3c5';
        mixm6to6_inr3c6    : Result := 'inr3c6';
        mixm6to6_inr4c1    : Result := 'inr4c1';
        mixm6to6_inr4c2    : Result := 'inr4c2';
        mixm6to6_inr4c3    : Result := 'inr4c3';
        mixm6to6_inr4c4    : Result := 'inr4c4';
        mixm6to6_inr4c5    : Result := 'inr4c5';
        mixm6to6_inr4c6    : Result := 'inr4c6';
        mixm6to6_inr5c1    : Result := 'inr5c1';
        mixm6to6_inr5c2    : Result := 'inr5c2';
        mixm6to6_inr5c3    : Result := 'inr5c3';
        mixm6to6_inr5c4    : Result := 'inr5c4';
        mixm6to6_inr5c5    : Result := 'inr5c5';
        mixm6to6_inr5c6    : Result := 'inr5c6';
        mixm6to6_inr6c1    : Result := 'inr6c1';
        mixm6to6_inr6c2    : Result := 'inr6c2';
        mixm6to6_inr6c3    : Result := 'inr6c3';
        mixm6to6_inr6c4    : Result := 'inr6c4';
        mixm6to6_inr6c5    : Result := 'inr6c5';
        mixm6to6_inr6c6    : Result := 'inr6c6';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModMixM6to6.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm6to6_out1      : Result := 'out1';
        mixm6to6_out2      : Result := 'out2';
        mixm6to6_out3      : Result := 'out3';
        mixm6to6_out4      : Result := 'out4';
        mixm6to6_out5      : Result := 'out5';
        mixm6to6_out6      : Result := 'out6';
        mixm6to6_outchrow1 : Result := 'outchrow1';
        mixm6to6_outchrow2 : Result := 'outchrow2';
        mixm6to6_outchrow3 : Result := 'outchrow3';
        mixm6to6_outchrow4 : Result := 'outchrow4';
        mixm6to6_outchrow5 : Result := 'outchrow5';
        mixm6to6_outchrow6 : Result := 'outchrow6';
        else                 Result := NO_OUTPUT;
      end;
    end;




{ ========
  TModQuantize = class( TMod)
  public
}

const

  quantize_qsteps = 0;
  quantize_in     = 1;
  quantize_mute   = 2;

  quantize_out    = 0;


    procedure   TModQuantize.DoTick; // override;
    var
      Steps: Integer;
    begin
      Steps := Max( 1, Round( FInCache[ quantize_qsteps]));
      FOutCache[ quantize_out] := Round( Steps * FInCache[ quantize_in] * SignalToMute( FInCache[ quantize_mute] )) / Steps;
    end;

    function    TModQuantize.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        quantize_qsteps : Result := 'qsteps';
        quantize_in     : Result := 'in';
        quantize_mute   : Result := 'mute';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModQuantize.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        quantize_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDisplay = class( TMod)
  private
    FMax1: TSignal;
    FMax2: TSignal;
  public
}

const

  display_in1  = 0;
  display_in2  = 1;
  display_mode = 2;


    constructor TModDisplay.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
    end;

    procedure   TModDisplay.DoSlowTick; // override;
    begin
      if SignalToInt( FInCache[ display_mode]) = 0
      then begin
        FMax1 := Max( FMax1, Abs( FInCache[ display_in1]));
        FMax2 := Max( FMax2, Abs( FInCache[ display_in2]));
      end
      else begin
        FMax1 := Max( FMax1, FInCache[ display_in1]);
        FMax2 := Max( FMax2, FInCache[ display_in2]);
      end;
    end;

    procedure   TModDisplay.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    const
      LOWVAL  = -45.0;
      HIGHVAL =  18.0;
      LOWLED  =   0.0;
      HIGHLED =   1.0;
    var
      adB1 : TSignal;
      adB2 : TSignal;
    begin
      adB1 := RangeMap( Clip( SignalTodB( FMax1), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      adB2 := RangeMap( Clip( SignalTodB( FMax2), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      aCallback( Self, Format( '%s/%s/level1/%f', [ aPrefix, Name, adB1], AppLocale));
      aCallback( Self, Format( '%s/%s/level2/%f', [ aPrefix, Name, adB2], AppLocale));
      FMax1 := 0;
      FMax2 := 0;
    end;

    function    TModDisplay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        display_in1  : Result := 'in1';
        display_in2  : Result := 'in2';
        display_mode : Result := 'mode';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModDisplay.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;



{ ========
  TModStateSpace = class( TMod)
  private
    FA        : TSignalMatrix;
    FB        : TSignalMatrix;
    FC        : TSignalMatrix;
    FD        : TSignalMatrix;
    FX        : TSignalMatrix;
    FNewX     : TSignalMatrix;
    FOut      : TSignalMatrix;
    FIn       : TSignalMatrix;
    FRunnable : Boolean;
  private
}

    function    TModStateSpace.Consistent: Boolean;
    begin
      Result :=
        FA.IsSquare                     and
        ( FA   .RowCount = FX.RowCount) and
        ( FX   .ColCount = 1          ) and
        ( FNewX.RowCount = FX.RowCount) and
        ( FNewX.ColCount = FX.ColCount) and
        ( FB   .RowCount = FX.RowCount) and
        ( FB   .ColCount = InputCount ) and
        ( FC   .RowCount = OutputCount) and
        ( FC   .ColCount = FX.ColCount) and
        ( FD   .RowCount = OutputCount) and
        ( FD   .ColCount = InputCount );
    end;

    procedure   TModStateSpace.ClearState;
    begin
      FX   .RowCount := FA.RowCount;
      FX   .ColCount := 1;
      FNewX.RowCount := FA.RowCount;
      FNewX.ColCount := 1;
      FX   .Clear;
      FNewX.Clear;
    end;

//  protected

    procedure   TModStateSpace.Run;
    var
      Tmp : TSignalMatrix;
    begin
      Tmp := TSignalMatrix.Create( FInCache);
      try
        FIn.Assign( Tmp);
        FIn.AntiRunaway;

        Tmp.Assign( FB);                          // FNewX( z) = FA.FX(z) + FB.FIn( z)
        Tmp.AntiRunaway;
        Tmp.Mult( FIn);
        Tmp.AntiRunaway;
        FNewX.Assign( FA);
        FNewX.AntiRunaway;
        FNewX.Mult( FX);
        FNewX.AntiRunaway;
        FNewX.Add( Tmp);
        FNewX.AntiRunaway;

        Tmp.Assign( FD);                          // FOut( z) = FC.FX( z) + FD.FIn( z)
        Tmp.AntiRunaway;
        Tmp.Mult( FIn);
        Tmp.AntiRunaway;
        FOut.Assign( FC);
        FOut.AntiRunaway;
        FOut.Mult( FX);
        FOut.AntiRunaway;
        FOut.Add( Tmp);
        FOut.AntiRunaway;

        FOut.PopulateVector( FOutCache);

        FX.Assign( FNewX);
      finally
        Tmp.Free;
      end;
    end;

//  public

    constructor TModStateSpace.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIn   := TSignalMatrix.Create;
      FOut  := TSignalMatrix.Create;
      FA    := TSignalMatrix.Create;
      FB    := TSignalMatrix.Create;
      FC    := TSignalMatrix.Create;
      FD    := TSignalMatrix.Create;
      FX    := TSignalMatrix.Create;
      FNewX := TSignalMatrix.Create;
      Setup;
      ClearState;
      FRunnable := Consistent;
    end;

    destructor  TModStateSpace.Destroy; // override;
    begin
      FRunnable := False;
      FreeAndNil( FNewX);
      FreeAndNil( FX   );
      FreeAndNil( FD   );
      FreeAndNil( FC   );
      FreeAndNil( FB   );
      FreeAndNil( FA   );
      FreeAndNil( FOut );
      FreeAndNil( FIn  );
      inherited;
    end;

    procedure   TModStateSpace.DoTick; // override;
    begin
      if FRunnable
      then Run;
    end;

    function    TModStateSpace.InputName( anIndex: Integer): string; // override;
    begin
      Result := NO_INPUT;
    end;

    function    TModStateSpace.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;


{ ========
  TModMidiIn = class( TMod)
  public
}

    constructor TModMidiIn.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsMidi := True;
    end;


{ ========
  TModMidiNoteIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FLastNote    : Byte;
    FLastChannel : Byte;
    FOldChannel  : Byte;
    FGate        : Boolean;
    FNote        : TSignal;
    FVelocityOn  : TSignal;
    FVelocityOff : TSignal;
  public
}

const

  midinotein_ch     = 0;

  midinotein_gate   = 0;
  midinotein_note   = 1;
  midinotein_velon  = 2;
  midinotein_veloff = 3;
{ midinotein_bend   = 4; }


    procedure   TModMidiNoteIn.NotesOff;
    begin
      FGate        := False;
      FVelocityOn  := 0;
      FVelocityOff := 0;
    end;

    procedure   TModMidiNoteIn.AcceptMidi( const aMsg: TMidiMessage); // override;
    begin
      if ( FChannel = 255) or ( aMsg.Channel = FChannel)            // 255 being omni
      then begin
        if Length( aMsg.Data) = 2
        then begin
          if aMSg.Command = CNoteOn
          then begin
            FGate        := True;
            FLastNote    := aMsg.Data[ 0];
            FLastChannel := aMSg.Channel;
            FNote        := ( FLastNote - MiddleNote) / 128;
            FVelocityOn  := aMsg.Data[ 1] / 128;
          end
          else if ( aMsg.Command = CNoteOff) and ( aMsg.Data[ 0] = FLastNote) and ( aMsg.Channel = FLastChannel)
          then begin
            FGate        := False;
            FVelocityOff := aMsg.Data[ 1] / 128;
          end
          else if ( aMsg.Command = CControlChange) and ( aMsg.Data[ 0] = CCAllNotesOff) and ( aMsg.Data[ 1] = 0) // All notes off
          then NotesOff;
        end;
      end;
    end;

    procedure   TModMidiNoteIn.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        NotesOff;
        ResetFlag := False;
      end;

      FChannel := SignalToInt( Clip( FInCache[ midinotein_ch], 0, 16));
      if FChannel = 0
      then FChannel := 255
      else FChannel := FChannel - 1;
      if FOldChannel <> FChannel
      then NotesOff;
      FOutCache[ midinotein_gate  ] := LogicToSignal( FGate);
      FOutCache[ midinotein_note  ] := FNote;
      FOutCache[ midinotein_velon ] := FVelocityOn;
      FOutCache[ midinotein_veloff] := FVelocityOff;
      FOldChannel   := FChannel;
    end;

    procedure   TModMidiNoteIn.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiNoteIn.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midinotein_ch : Result := 'ch';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModMidiNoteIn.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midinotein_gate   : Result := 'gate';
        midinotein_note   : Result := 'note';
        midinotein_velon  : Result := 'velon';
        midinotein_veloff : Result := 'veloff';
     // midinotein_bend   : Result := 'bend';
        else                Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiCCIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FCC          : Byte;
    FGate        : Boolean;
    FValue       : TSignal;
    FCounter     : Integer;
    FLedCounter  : Integer;
  public
}

const

  midiccin_ch    = 0;
  midiccin_cc    = 1;

  midiccin_gate  = 0;
  midiccin_value = 1;


    procedure   TModMidiCCIn.AcceptMidi( const aMsg: TMidiMessage); // override;
    begin
      if ( FChannel = 255) or ( aMsg.Channel = FChannel)            // 255 being omni
      then begin
        if Length( aMsg.Data) = 2
        then begin
          if ( aMsg.Command = CControlChange) and ( aMsg.Data[ 0] = FCC)
          then begin
            FValue      := aMsg.Data[ 1];
            FCounter    := 44;                 //   1 ms gate pulse
            FLedCounter := 11025;              // 250 ms LED  pulse
            FGate       := True;
          end;
        end;
      end;
    end;

    procedure   TModMidiCCIn.DoSlowTick; // override;
    begin
      if FLedCounter > 0
      then Dec( FLedCounter);
      FChannel := SignalToInt( FInCache[ midiccin_ch]);
      FCC      := SignalToInt( FInCache[ midiccin_cc]);
      if FCounter > 0
      then Dec( FCounter)
      else FGate := False;
      FOutCache[ midiccin_gate ] := LogicToSignal( FGate);
      FOutCache[ midiccin_value] := FValue / 128;
    end;

    procedure   TModMidiCCIn.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FLedCounter <> 0)], AppLocale));
    end;

    function    TModMidiCCIn.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiccin_ch : Result := 'ch';
        midiccin_cc : Result := 'cc';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMidiCCIn.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiccin_gate  : Result := 'gate';
        midiccin_value : Result := 'value';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiOut = class( TMod)
  public
}

    constructor TModMidiOut.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsMidi := False;
    end;


{ ========
  TModMidiNoteOut = class( TModMidiOut)
  private
    FLastNote    : Byte;
    FLastChannel : Byte;
    FGate        : Boolean;
    FDidSend     : Boolean;
  public
}

const

  midinoteout_ch     = 0;
  midinoteout_gate   = 1;
  midinoteout_note   = 2;
  midinoteout_velon  = 3;
  midinoteout_veloff = 4;


    procedure   TModMidiNoteOut.SetDefaults; // override;
    begin
      FInCache[ midinoteout_gate  ] := 0;    // Gate off
      FInCache[ midinoteout_note  ] := 0;    // Note 0
      FInCache[ midinoteout_velon ] := 1;    // Max velon
      FInCache[ midinoteout_veloff] := 1;    // Max veloff
    end;

    procedure   TModMidiNoteOut.DoSlowTick; // override;
    // FNote := ( FLastNote - 69) / 128;
    var
      NewGate : Boolean;
      aMsg    : TMidiMessage;
    begin
      NewGate := SignalToLogic( FInCache[ midinoteout_gate]);
      if NewGate <> FGate
      then begin
        if NewGate
        then begin
          FLastChannel := SignalToInt( FinCache[ midinoteout_ch]);
          FLastNote    := SignalToInt( Clip( FInCache[ midinoteout_note] * 128 + MiddleNote, 0, 127));
          SetLength( aMsg.Data, 2);
          aMsg.Command  := CNoteOn;
          aMsg.Channel  := FLastChannel - 1;
          aMsg.Data[ 0] := FlastNote;
          aMsg.Data[ 1] := SignalToInt( Clip( FInCache[ midinoteout_velon], 0, 1) * 127);
          SendMidi( aMsg);
          FDidSend := True;
        end
        else begin
          if FDidSend
          then begin
            FDidSend := False;
            SetLength( aMsg.Data, 2);
            aMsg.Command  := CNoteOff;
            aMsg.Channel  := FLastChannel - 1;
            aMsg.Data[ 0] := FlastNote;
            aMsg.Data[ 1] := SignalToInt( Clip( FInCache[ midinoteout_veloff], 0, 1) * 127);
            SendMidi( aMsg);
          end;
        end;
        FGate := NewGate;
      end;
    end;

    procedure   TModMidiNoteOut.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiNoteOut.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midinoteout_ch     : Result := 'ch';
        midinoteout_gate   : Result := 'gate';
        midinoteout_note   : Result := 'note';
        midinoteout_velon  : Result := 'velon';
        midinoteout_veloff : Result := 'veloff';
        else                 Result := NO_INPUT;
      end;
    end;


{ ========
  TModMidiCCOut = class( TModMidiOut)
  private
    FGate  : Boolean;
  public
}

const

  midiccout_ch    = 0;
  midiccout_cc    = 1;
  midiccout_gate  = 2;
  midiccout_value = 3;


    procedure   TModMidiCCOut.DoSlowTick; // override;
    var
      NewGate : Boolean;
      aMsg    : TMidiMessage;
    begin
      NewGate := SignalToLogic( FInCache[ midiccout_gate]);
      if NewGate <> FGate
      then begin
        if NewGate
        then begin
          SetLength( aMsg.Data, 2);
          aMsg.Command  := CControlChange;
          aMsg.Channel  := SignalToInt( FInCache[ midiccout_ch]);
          aMsg.Data[ 0] := SignalToInt( FInCache[ midiccout_cc]);
          aMsg.Data[ 1] := SignalToInt( Clip( FInCache[ midiccout_value], 0, 1) * 127);
          SendMidi( aMsg);
        end;
        FGate := NewGate;
      end;
    end;

    procedure   TModMidiCCOut.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiCCOut.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiccout_ch    : Result := 'ch';
        midiccout_cc    : Result := 'cc';
        midiccout_gate  : Result := 'gate';
        midiccout_value : Result := 'value';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModMidiCCOut.OutputName( anIndex: Integer): string; // override;
    begin
      Result := inherited;
    end;


end.
